# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2024 Sarah Hoffmann.
import osmium as o

import pytest

from helpers import IDCollector

def test_filter_no_keys():
    with pytest.raises(TypeError, match="keys to filter"):
        o.filter.KeyFilter()


@pytest.mark.parametrize('key', [None, 1, IDCollector(), 'a'.encode('utf-8')])
def test_filter_bad_argument_types(key):
    with pytest.raises(TypeError, match="must be strings"):
        o.filter.KeyFilter("foo", key)


@pytest.mark.parametrize('key,nodes,changesets', [('foo', [1], [10]),
                                                  ('name', [1, 2], [20]),
                                                  ('kö*', [4], [])])
def test_filter_simple(opl_reader, key, nodes, changesets):
    data = """\
           n1 Tfoo=bar,name=loo
           n2 Tname=else
           n3 x9 y0
           n4 Tkö*=fr
           c10 Tfoo=baz
           c20 Tname=none
           """

    post = IDCollector()

    o.apply(opl_reader(data), o.filter.KeyFilter(key), post)

    assert post.nodes == nodes
    assert post.changesets == changesets



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest

import osmium as o

from helpers import IDCollector

def test_simple_way(test_data, tmp_path):
    ref_file = test_data('\n'.join((f"n{i} x2 y3" for i in range(10))))

    class TestWay:
        id = 34
        nodes = [3, 6, 5]

    outfile = str(tmp_path / 'test.osm')

    with o.BackReferenceWriter(outfile, ref_file) as writer:
        writer.add_way(TestWay())

    ids = IDCollector()
    o.apply(outfile, ids)

    assert ids.nodes == [3, 5, 6]
    assert ids.ways == [34]
    assert not ids.relations


def test_do_not_write_on_exception(test_data, tmp_path):
    ref_file = test_data('\n'.join((f"n{i} x2 y3" for i in range(10))))
    outfile = tmp_path / 'test.osm'

    with pytest.raises(RuntimeError, match="inner error"):
        with o.BackReferenceWriter(str(outfile), ref_file) as writer:
            raise RuntimeError("inner error")

    assert not outfile.exists()



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2024 Sarah Hoffmann.
import osmium as o

import pytest

from helpers import IDCollector

@pytest.fixture
def reader(opl_reader):
    return opl_reader("""\
               n1 x1 y1
               n2 x1 y1 Tfoo=bar
               w1 Nn1,n2 Thighway=road
               w34 Nn2,n1
               r90
               r91 Tsome=thing
               c222 Ttodo=done
               c223
               """)

def test_filter_default_config(reader):
    pre = IDCollector()
    post = IDCollector()
    o.apply(reader, pre, o.filter.EmptyTagFilter(), post)

    assert pre.nodes == [1, 2]
    assert post.nodes == [2]
    assert pre.ways == [1, 34]
    assert post.ways == [1]
    assert pre.relations == [90, 91]
    assert post.relations == [91]
    assert pre.changesets == [222, 223]
    assert post.changesets == [222]


def test_filter_restrict_entity(reader):
    pre = IDCollector()
    post = IDCollector()
    o.apply(reader, pre, o.filter.EmptyTagFilter().enable_for(o.osm.WAY | o.osm.RELATION), post)

    assert pre.nodes == [1, 2]
    assert post.nodes == [1, 2]
    assert pre.ways == [1, 34]
    assert post.ways == [1]
    assert pre.relations == [90, 91]
    assert post.relations == [91]


def test_filter_chained(reader):
    pre = IDCollector()
    post = IDCollector()
    o.apply(reader, pre,
            o.filter.EmptyTagFilter().enable_for(o.osm.NODE),
            o.filter.EmptyTagFilter().enable_for(o.osm.WAY),
            post)

    assert pre.nodes == [1, 2]
    assert post.nodes == [2]
    assert pre.ways == [1, 34]
    assert post.ways == [1]
    assert pre.relations == [90, 91]
    assert post.relations == [90, 91]



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2022 Sarah Hoffmann.
import pytest

import osmium as o

@pytest.fixture
def tag_handler(simple_handler):

    def _handle(data, tests):
        tags = {None: None} # marker that handler hasn't run yet
        def node(n):
            if None in tags:
                del tags[None]
            tags.update(n.tags)
            tests(n)

        simple_handler(data, node=node)

        return tags

    return _handle


def test_empty_taglist_length(tag_handler):
    def tests(n):
        assert 0 == len(n.tags)
        assert not n.tags
        assert str(n.tags) == '{}'
        assert repr(n.tags) == 'osmium.osm.TagList({})'

    tags = tag_handler("n234 x1 y2", tests)
    assert tags == {}


def test_empty_taglist_contains(tag_handler):
    def tests(n):
        assert "a" not in n.tags

    tags = tag_handler("n234 x1 y2", tests)
    assert tags == {}


def test_empty_taglist_get(tag_handler):
    def tests(n):
        assert None == n.tags.get("foo")
        assert None == n.tags.get("foo", None)
        assert "fs" == n.tags.get("foo", "fs")

    tags = tag_handler("n234 x1 y2", tests)
    assert tags == {}


def test_empty_taglist_indexop(tag_handler):
    def tests(n):
        with pytest.raises(KeyError):
            n.tags["foo"]
        with pytest.raises(KeyError):
            n.tags[None]

    tags = tag_handler("n234 x1 y2", tests)
    assert tags == {}


def test_taglist_length(simple_handler):
    data = u"""\
           n1 x0 y0 Ta=a
           n2 Tkeyñ=value
           n3 Tfoo=1ß,bar=2,foobar=33
           """

    lens = {}
    def node(n):
        lens[n.id] = len(n.tags)
        assert n.tags

    simple_handler(data, node=node)

    lens = {1 : 1, 2 : 1, 3 : 3}


def test_taglist_contains(tag_handler):
    def tests(n):
        assert "abba" in n.tags
        assert "2" in n.tags
        assert "xx" in n.tags
        assert "x" not in n.tags
        assert None not in n.tags
        assert "" not in n.tags
        assert str(n.tags) == '{abba=x,2=vvv,xx=abba}'
        assert repr(n.tags) == "osmium.osm.TagList({'abba': 'x', '2': 'vvv', 'xx': 'abba'})"

    tags = tag_handler("n234 Tabba=x,2=vvv,xx=abba", tests)

    assert tags == {'abba': 'x', '2': 'vvv', 'xx': 'abba'}


def test_taglist_indexop(tag_handler):
    def tests(n):
        assert "x" == n.tags["abba"]
        assert "vvv" == n.tags["2"]
        assert "abba" == n.tags["xx"]
        for k in ("x", "addad", "..", None):
            with pytest.raises(KeyError):
                n.tags[k]

    tags = tag_handler("n234 Tabba=x,2=vvv,xx=abba", tests)

    assert tags == {'abba': 'x', '2': 'vvv', 'xx': 'abba'}


def test_taglist_indexop(tag_handler):
    def tests(n):
        assert "x" == n.tags.get("abba")
        assert "vvv" == n.tags.get("2", None)
        assert "abba" == n.tags.get("xx", "ff")
        assert "43 fg" == n.tags.get("_", "43 fg")
        assert n.tags.get("gerger4") is None
        assert n.tags.get("ffleo", None) is None

    tags = tag_handler("n234 Tabba=x,2=vvv,xx=abba", tests)

    assert tags == {'abba': 'x', '2': 'vvv', 'xx': 'abba'}



""" Provides some helper functions for test.
"""
from datetime import datetime, timezone

import osmium

def mkdate(*args):
    return datetime(*args, tzinfo=timezone.utc)

class CountingHandler(osmium.SimpleHandler):

    def __init__(self):
        super(CountingHandler, self).__init__()
        self.counts = [0, 0, 0, 0]

    def node(self, _):
        self.counts[0] += 1

    def way(self, _):
        self.counts[1] += 1

    def relation(self, _):
        self.counts[2] += 1

    def area(self, _):
        self.counts[3] += 1


class IDCollector:

    def __init__(self):
        self.nodes = []
        self.ways = []
        self.relations = []
        self.changesets = []

    def node(self, n):
        self.nodes.append(n.id)

    def way(self, w):
        self.ways.append(w.id)

    def relation(self, r):
        self.relations.append(r.id)

    def changeset(self, c):
        self.changesets.append(c.id)



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2022 Sarah Hoffmann.
from contextlib import contextmanager
from collections import OrderedDict
from datetime import datetime, timezone, timedelta
import uuid

import pytest

import osmium as o

from helpers import mkdate


@pytest.fixture
def test_writer(tmp_path):
    @contextmanager
    def _WriteExpect(filename, expected):
        with o.SimpleWriter(str(filename), 1024*1024) as writer:
            yield writer

        assert filename.read_text().strip() == expected


    def _create(expected):
        filename = tmp_path / (str(uuid.uuid4()) + '.opl')
        return _WriteExpect(filename, expected)

    return _create


class O:
    def __init__(self, **params):
        for k,v in params.items():
            setattr(self, k, v)


@pytest.mark.parametrize('osmobj, out_attr', [
      (O(id=None), '0 v0 dV c0 t i0 u T'),
      (O(visible=None), '0 v0 dV c0 t i0 u T'),
      (O(version=None), '0 v0 dV c0 t i0 u T'),
      (O(uid=None), '0 v0 dV c0 t i0 u T'),
      (O(user=None), '0 v0 dV c0 t i0 u T'),
      (O(timestamp=None), '0 v0 dV c0 t i0 u T'),
      (O(id=1), '1 v0 dV c0 t i0 u T'),
      (O(id=-99), '-99 v0 dV c0 t i0 u T'),
      (O(visible=True), '0 v0 dV c0 t i0 u T'),
      (O(visible=False), '0 v0 dD c0 t i0 u T'),
      (O(version=23), '0 v23 dV c0 t i0 u T'),
      (O(user="Schmidt"), '0 v0 dV c0 t i0 uSchmidt T'),
      (O(user=""), '0 v0 dV c0 t i0 u T'),
      (O(uid=987), '0 v0 dV c0 t i987 u T'),
      (O(timestamp='2012-04-14T20:58:35Z'), '0 v0 dV c0 t2012-04-14T20:58:35Z i0 u T'),
      (O(timestamp=mkdate(2009, 4, 14, 20, 58, 35)), '0 v0 dV c0 t2009-04-14T20:58:35Z i0 u T'),
      (O(timestamp=datetime(2009, 4, 14, 20, 58, 35, tzinfo=timezone(timedelta(hours=1)))), '0 v0 dV c0 t2009-04-14T19:58:35Z i0 u T'),
      (O(timestamp='1970-01-01T00:00:01Z'), '0 v0 dV c0 t1970-01-01T00:00:01Z i0 u T')
    ])
class TestWriteAttributes:
    def test_node_simple_attr(self, test_writer, osmobj, out_attr):
        with test_writer('n' + out_attr + ' x y') as w:
            w.add_node(osmobj)

    def test_way_simple_attr(self, test_writer, osmobj, out_attr):
        with test_writer('w' + out_attr + ' N') as w:
            w.add_way(osmobj)

    def test_relation_simple_attr(self, test_writer, osmobj, out_attr):
        with test_writer('r' + out_attr + ' M') as w:
            w.add_relation(osmobj)


@pytest.mark.parametrize('tags,out', [
     (None, 'T'),
     ([], 'T'),
     ({}, 'T'),
     ((("foo", "bar"), ), 'Tfoo=bar'),
     ((("foo", "bar"), ("2", "1")), 'Tfoo=bar,2=1'),
     ({'test' : 'drive'}, 'Ttest=drive'),
     (OrderedDict((('a', 'b'), ('c', '3'))), 'Ta=b,c=3'),
    ])
class TestWriteTags:
    def test_node_tags(self, test_writer, tags, out):
        with test_writer('n0 v0 dV c0 t i0 u ' + out + ' x y') as w:
            w.add_node(O(tags=tags))

    def test_way_tags(self, test_writer, tags, out):
        with test_writer('w0 v0 dV c0 t i0 u ' + out + ' N') as w:
            w.add_way(O(tags=tags))

    def test_relation_tags(self, test_writer, tags, out):
        with test_writer('r0 v0 dV c0 t i0 u ' + out + ' M') as w:
            w.add_relation(O(tags=tags))


@pytest.mark.parametrize("location,out", [((1.1234561, 0.1234561), 'x1.1234561 y0.1234561'),
                                          ((30.46, 50.37), 'x30.46 y50.37'),
                                          (None, 'x y')])
def test_location(test_writer, location, out):
    with test_writer('n0 v0 dV c0 t i0 u T ' + out) as w:
        w.add_node(O(location=location))


def test_location_generic(test_writer):
    with test_writer('n0 v0 dV c0 t i0 u T x30.46 y50.37') as w:
        w.add_node(O(location=(30.46, 50.37)))


def test_node_list(test_writer):
    with test_writer('w0 v0 dV c0 t i0 u T Nn1,n2,n3,n-4') as w:
        w.add_way(O(nodes=(1, 2, 3, -4)))


def test_node_list_generic(test_writer):
    with test_writer('w0 v0 dV c0 t i0 u T Nn1,n2,n3,n-4') as w:
        w.add(O(nodes=(1, 2, 3, -4)))


def test_node_list_none(test_writer):
    with test_writer('w0 v0 dV c0 t i0 u T N') as w:
        w.add_way(O(nodes=None))


def test_relation_members(test_writer):
    with test_writer('r0 v0 dV c0 t i0 u T Mn34@foo,r200@,w1111@x') as w:
        w.add_relation(O(members=(('n', 34, 'foo'),
                                  ('r', 200, ''),
                                  ('w', 1111, 'x')
                                 )))


def test_relation_members_generic(test_writer):
    with test_writer('r0 v0 dV c0 t i0 u T Mn34@foo,r200@,w1111@x') as w:
        w.add(O(members=(('n', 34, 'foo'),
                         ('r', 200, ''),
                         ('w', 1111, 'x')
                        )))


def test_relation_members_None(test_writer):
    with test_writer('r0 v0 dV c0 t i0 u T M') as w:
        w.add_relation(O(members=None))


def test_node_object(test_writer, simple_handler):
    node_opl = 'n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45'

    with test_writer(node_opl) as w:
        simple_handler(node_opl, node=lambda o: w.add_node(o))


def test_node_object_generic(test_writer, simple_handler):
    node_opl = 'n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45'

    with test_writer(node_opl) as w:
        simple_handler(node_opl, node=lambda o: w.add(o))


def test_location_object(test_writer, simple_handler):
    node_opl = 'n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45'

    with test_writer('n0 v0 dV c0 t i0 u T x98.7 y-3.45') as w:
        simple_handler(node_opl, node=lambda o: w.add_node(O(location=o.location)))


def test_tag_object(test_writer, simple_handler):
    node_opl = 'n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45'

    with test_writer('n0 v0 dV c0 t i0 u Telephant=yes x y') as w:
        simple_handler(node_opl, node=lambda o: w.add_node(O(tags=o.tags)))


def test_way_object(test_writer, simple_handler):
    way_opl = 'w45 v14 dV c0 t i0 u Thighway=top Nn23,n56,n34,n23'

    with test_writer(way_opl) as w:
        simple_handler(way_opl, way=lambda o: w.add_way(o))


def test_way_object_generic(test_writer, simple_handler):
    way_opl = 'w45 v14 dV c0 t i0 u Thighway=top Nn23,n56,n34,n23'

    with test_writer(way_opl) as w:
        simple_handler(way_opl, way=lambda o: w.add(o))


def test_nodelist_object(test_writer, simple_handler):
    way_opl = 'w45 v14 dV c0 t i0 u Thighway=top Nn23,n56,n34,n23'

    with test_writer('w0 v0 dV c0 t i0 u T Nn23,n56,n34,n23') as w:
        simple_handler(way_opl, way=lambda o: w.add_way(O(nodes=o.nodes)))


def test_noderef_object(test_writer, simple_handler):
    way_opl = 'w45 v14 dV c0 t i0 u Thighway=top Nn23,n56,n34,n23'

    with test_writer('w0 v0 dV c0 t i0 u T Nn56,n34') as w:
        simple_handler(way_opl,
                       way=lambda o: w.add_way(O(nodes=[n for n in o.nodes if n.ref != 23])))


def test_relation_object(test_writer, simple_handler):
    rel_opl = 'r2 v0 dV c0 t i0 u Ttype=multipolygon Mw1@,w2@,w3@inner'

    with test_writer(rel_opl) as w:
        simple_handler(rel_opl, relation=lambda o: w.add_relation(o))


def test_relation_object_generic(test_writer, simple_handler):
    rel_opl = 'r2 v0 dV c0 t i0 u Ttype=multipolygon Mw1@,w2@,w3@inner'

    with test_writer(rel_opl) as w:
        simple_handler(rel_opl, relation=lambda o: w.add(o))


def test_memberlist_object(test_writer, simple_handler):
    rel_opl = 'r2 v0 dV c0 t i0 u Ttype=multipolygon Mw1@,w2@,w3@inner'

    with test_writer('r0 v0 dV c0 t i0 u T Mw1@,w2@,w3@inner') as w:
        simple_handler(rel_opl,
                       relation=lambda o: w.add_relation(O(members=o.members)))


def test_member_object(test_writer, simple_handler):
    rel_opl = 'r2 v0 dV c0 t i0 u Ttype=multipolygon Mw1@,w2@,w3@inner'

    with test_writer('r0 v0 dV c0 t i0 u T Mw1@,w2@') as w:
        simple_handler(rel_opl,
                       relation=lambda o: w.add_relation(O(members=[m for m in o.members
                                                                    if m.role != 'inner'])))


def test_set_custom_header(tmp_path):
    fn = str(tmp_path / 'test.xml')
    h = o.io.Header()
    h.set('generator', 'foo')
    h.add_box(o.osm.Box(0.1, -4, 10, 45))

    writer = o.SimpleWriter(fn, 4000, h)

    try:
        writer.add_node({})
    finally:
        writer.close()

    with o.io.Reader(fn) as rd:
        h = rd.header()
        assert h.get('generator') == 'foo'
        assert h.box().valid()
        assert h.box().bottom_left == o.osm.Location(0.1, -4)
        assert h.box().top_right == o.osm.Location(10, 45)


def test_add_node_after_close(tmp_path, simple_handler):
    node_opl = "n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45"

    filename = tmp_path / (str(uuid.uuid4()) + '.opl')
    writer = o.SimpleWriter(str(filename), 1024*1024)
    writer.close()

    with pytest.raises(RuntimeError, match='closed'):
        simple_handler(node_opl, node=lambda o: writer.add_node(o))


def test_add_way_after_close(tmp_path, simple_handler):
    node_opl = "w1 Nn1"

    filename = tmp_path / (str(uuid.uuid4()) + '.opl')
    writer = o.SimpleWriter(str(filename), 1024*1024)
    writer.close()

    with pytest.raises(RuntimeError, match='closed'):
        simple_handler(node_opl, way=lambda o: writer.add_way(o))

def test_add_relation_after_close(tmp_path, simple_handler):
    node_opl = "r54 Mn1@,w3@foo"

    filename = tmp_path / (str(uuid.uuid4()) + '.opl')
    writer = o.SimpleWriter(str(filename), 1024*1024)
    writer.close()

    with pytest.raises(RuntimeError, match='closed'):
        simple_handler(node_opl, relation=lambda o: writer.add_relation(o))


@pytest.mark.parametrize("final_item", (True, False))
def test_catch_errors_in_add_node(tmp_path, final_item):
    test_file = tmp_path / 'test.opl'

    with o.SimpleWriter(str(test_file), 4000) as writer:
        writer.add_node(o.osm.mutable.Node(id=123))
        with pytest.raises(TypeError):
            writer.add_node(o.osm.mutable.Node(id=124, tags=34))
        if not final_item:
            writer.add_node(o.osm.mutable.Node(id=125))

    output = test_file.read_text()

    expected = 'n123 v0 dV c0 t i0 u T x y\n'
    if not final_item:
        expected += 'n125 v0 dV c0 t i0 u T x y\n'

    assert output == expected


@pytest.mark.parametrize("final_item", (True, False))
def test_catch_errors_in_add_way(tmp_path, final_item):
    test_file = tmp_path / 'test.opl'

    with o.SimpleWriter(test_file, 4000) as writer:
        writer.add_way(o.osm.mutable.Way(id=123, nodes=[1, 2, 3]))
        with pytest.raises(TypeError):
            writer.add_way(o.osm.mutable.Way(id=124, nodes=34))
        if not final_item:
            writer.add_way(o.osm.mutable.Way(id=125, nodes=[11, 12]))

    output = test_file.read_text()

    expected = 'w123 v0 dV c0 t i0 u T Nn1,n2,n3\n'
    if not final_item:
        expected += 'w125 v0 dV c0 t i0 u T Nn11,n12\n'

    assert output == expected


@pytest.mark.parametrize("final_item", (True, False))
def test_catch_errors_in_add_relation(tmp_path, final_item):
    test_file = tmp_path / 'test.opl'

    with o.SimpleWriter(filename=str(test_file), bufsz=4000) as writer:
        writer.add_relation(o.osm.mutable.Relation(id=123))
        with pytest.raises(TypeError):
            writer.add_relation(o.osm.mutable.Relation(id=124, members=34))
        if not final_item:
            writer.add_relation(o.osm.mutable.Relation(id=125))

    output = test_file.read_text()

    expected = 'r123 v0 dV c0 t i0 u T M\n'
    if not final_item:
        expected += 'r125 v0 dV c0 t i0 u T M\n'

    assert output == expected


def test_do_not_overwrite_by_default(tmp_path):
    test_file = tmp_path / 'test.opl'

    with o.SimpleWriter(filename=str(test_file), bufsz=4000) as writer:
        writer.add_node(o.osm.mutable.Node(id=123))

    # try to open again
    with pytest.raises(RuntimeError, match='Open failed'):
        o.SimpleWriter(filename=str(test_file))


def test_do_overwrite(tmp_path):
    test_file = tmp_path / 'test.opl'

    with o.SimpleWriter(filename=str(test_file), bufsz=4000) as writer:
        writer.add_node(o.osm.mutable.Node(id=123))

    with o.SimpleWriter(filename=str(test_file), overwrite=True) as writer:
        pass



def test_write_to_file(tmp_path):
    test_file = tmp_path / 'test.txt'

    with o.SimpleWriter(o.io.File(test_file, 'opl'), bufsz=4000) as writer:
        writer.add_node(o.osm.mutable.Node(id=123))



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import Sequence, Any, NamedTuple, Callable, Optional, Iterator, \
                   Iterable, TYPE_CHECKING, TypeVar, Generic, Tuple, Union
import datetime as dt

import osmium.osm.mutable


if TYPE_CHECKING:
    import osmium.osm._osm as cosm

T_obj = TypeVar('T_obj', 'cosm.COSMNode', 'cosm.COSMWay', 'cosm.COSMRelation', 'cosm.COSMArea')

def _make_repr(name: str, *attrs: str) -> Callable[[Any], str]:
    fmt_string = f'osmium.osm.{name}('\
                 + ', '.join([f'{x}={{0.{x}!r}}' for x in attrs])\
                 + ')'

    def _repr(self: Any) -> str:
        if self._pyosmium_data.is_valid():
            return fmt_string.format(self)

        return f'osmium.osm.{name}(<invalid>)'

    return _repr


def _list_elipse(obj: Iterable[Any]) -> str:
    objects = ','.join(map(str, obj))
    if len(objects) > 50:
        objects = objects[:47] + '...'
    return objects


class Tag(NamedTuple):
    """ A single OSM tag.
    """

    k: str
    "Tag key"

    v: str
    "Tag value"

    def __str__(self) -> str:
        return f"{self.k}={self.v}"


class TagIterator:

    def __init__(self, parent: 'cosm.TagContainerProtocol') -> None:
        self._pyosmium_data = parent
        self.iterator = self._pyosmium_data.tags_begin()

    def __iter__(self) -> 'TagIterator':
        return self

    def __next__(self) -> Tag:
        return self._pyosmium_data.tags_next(self.iterator)


class TagList(Iterable[Tag]):
    """ A fixed list of tags. The list is exported as an unmutable,
        dictionary-like object where the keys are tag strings and the
        items are [Tags][osmium.osm.Tag].
    """

    def __init__(self, parent: 'cosm.TagContainerProtocol') -> None:
        self._pyosmium_data = parent

    def __len__(self) -> int:
        return self._pyosmium_data.tags_size()

    def __getitem__(self, key: str) -> str:
        if key is None:
            raise KeyError("Key 'None' not allowed.")

        val = self._pyosmium_data.tags_get_value_by_key(key, None)
        if val is None:
            raise KeyError("No tag with that key.")

        return val

    def __contains__(self, key: str) -> bool:
        return key is not None and self._pyosmium_data.tags_has_key(key)

    def get(self, key: str, default: Optional[str] = None) -> Optional[str]:
        """ Return the value for the given key. or 'value' if the key
            does not exist in the list.
        """
        if key is None:
            return default

        return self._pyosmium_data.tags_get_value_by_key(key, default)

    def __iter__(self) -> TagIterator:
        return TagIterator(self._pyosmium_data)

    def __str__(self) -> str:
        if self._pyosmium_data.is_valid():
            return f"{{{_list_elipse(self)}}}"

        return '<taglist invalid>'

    def __repr__(self) -> str:
        if self._pyosmium_data.is_valid():
            tagstr = ', '.join([f"{i.k!r}: {i.v!r}" for i in self])
        else:
            tagstr = '<invalid>'

        return f"osmium.osm.TagList({{{tagstr}}})"


class NodeRef:
    """ A reference to a OSM node that also caches the nodes location.
    """

    def __init__(self, location: 'osmium.osm.Location', ref: int) -> None:
        self.location = location
        self.ref = ref

    @property
    def x(self) -> int:
        """ (read-only) X coordinate (longitude) as a fixed-point integer.
        """
        return self.location.x

    @property
    def y(self) -> int:
        """ (read-only) Y coordinate (latitude) as a fixed-point integer.
        """
        return self.location.y

    @property
    def lat(self) -> float:
        """ (read-only) Latitude (y coordinate) as floating point number.
        """
        return self.location.lat

    @property
    def lon(self) -> float:
        """ (read-only) Longitude (x coordinate) as floating point number.
        """
        return self.location.lon

    def __str__(self) -> str:
        if self.location.valid():
            return f"{self.ref:d}@{self.location!s}"

        return str(self.ref)

    def __repr__(self) -> str:
        return f"osmium.osm.NodeRef(ref={self.ref!r}, location={self.location!r})"


class NodeRefList:
    """ A list of node references, implemented as
        an immutable sequence of [osmium.osm.NodeRef][]. This class
        is normally not used directly, use one of its subclasses instead.
    """

    def __init__(self, parent: 'cosm.BufferProxyProtocol', ref_list: 'cosm.NodeRefList') -> None:
        self._pyosmium_data = parent
        self._list = ref_list

    def is_closed(self) -> bool:
        """ True if the start and end node are the same (synonym for
            ``ends_have_same_id``).
        """
        return self._list.is_closed(self._pyosmium_data)

    def ends_have_same_id(self) -> bool:
        """ True if the start and end node are exactly the same.
        """
        return self._list.is_closed(self._pyosmium_data)

    def ends_have_same_location(self) -> bool:
        """ True if the start and end node of the way are at the same location. "
            Expects that the coordinates of the way nodes have been loaded
            ([SimpleHandler apply functions][osmium.SimpleHandler] and
            [`FileProcessor.with_locations()`][osmium.FileProcessor.with_locations]).
            If the locations are not present then the function returns always true.
        """
        return self._list.ends_have_same_location(self._pyosmium_data)

    def __len__(self) -> int:
        return self._list.size(self._pyosmium_data)

    def __getitem__(self, idx: int) -> NodeRef:
        return self._list.get(self._pyosmium_data, idx)

    def __iter__(self) -> Iterator[NodeRef]:
        return (self[i] for i in range(len(self)))

    def __str__(self) -> str:
        if self._pyosmium_data.is_valid():
            return f'[{_list_elipse(self)}]'

        return '<node list invalid>'

    def __repr__(self) -> str:
        if self._pyosmium_data.is_valid():
            nodes = ', '.join(map(repr, self))
        else:
            nodes = '<invalid>'

        return 'osmium.osm.{}([{}])'.format(self.__class__.__name__, nodes)


class WayNodeList(NodeRefList):
    """ List of nodes in a way.
        For its members see [`osmium.osm.NodeRefList`][].
    """


class OuterRing(NodeRefList):
    """List of nodes in an outer ring.
       For its members see [`osmium.osm.NodeRefList`][].
    """

class InnerRing(NodeRefList):
    """ List of nodes in an inner ring. "
        For its members see [`osmium.osm.NodeRefList`][].
    """

class RelationMember:
    """ Single member of a relation.
    """
    ref : int
    "OSM ID of the object. Only unique within the type."
    type: str
    "Type of object referenced, a node, way or relation."
    role: str
    """ The role of the member within the relation, a free-text string.
        If no role is set then the string is empty.
    """

    def __init__(self, ref: int, mtype: str, role: str):
        self.ref = ref
        self.type = mtype
        self.role = role

    def __str__(self) -> str:
        if self.role:
            return f"{self.type}{self.ref:d}@{self.role}"

        return f"{self.type}{self.ref:d}"

    def __repr__(self) -> str:
        return f"osmium.osm.RelationMember(ref={self.ref!r}, type={self.type!r}, role={self.role!r})"


class MemberIterator:
    """ Iterator over a sequence of RelationMembers.
    """

    def __init__(self, parent: 'cosm.COSMRelation') -> None:
        self._pyosmium_data = parent
        self.iterator = self._pyosmium_data.members_begin()

    def __iter__(self) -> 'MemberIterator':
        return self

    def __next__(self) -> RelationMember:
        return self._pyosmium_data.members_next(self.iterator)


class RelationMemberList:
    """ An immutable  sequence of relation members
        [`osmium.osm.RelationMember`][].
    """

    def __init__(self, parent: 'cosm.COSMRelation') -> None:
        self._pyosmium_data = parent

    def __len__(self) -> int:
        return self._pyosmium_data.members_size()

    def __iter__(self) -> MemberIterator:
        return MemberIterator(self._pyosmium_data)

    def __str__(self) -> str:
        if self._pyosmium_data.is_valid():
            return f'[{_list_elipse(self)}]'

        return '<member list invalid>'

    def __repr__(self) -> str:
        if self._pyosmium_data.is_valid():
            members = ', '.join(map(repr, self))
        else:
            members = '<invalid>'

        return f'osmium.osm.RelationMemberList([{members}])'


class OSMObject(Generic[T_obj]):
    """ This is the base class for all OSM entity classes below and contains
        all common attributes.
    """
    _pyosmium_data: T_obj
    _tags: TagList

    @property
    def id(self) -> int:
        """ (read-only) OSM id of the object.
        """
        return self._pyosmium_data.id()

    @property
    def deleted(self) -> bool:
        """ (read-only) True if the object is no longer visible.
        """
        return self._pyosmium_data.deleted()

    @property
    def visible(self) -> bool:
        """ (read-only) True if the object is visible.
        """
        return self._pyosmium_data.visible()

    @property
    def version(self) -> int:
        """ (read-only) Version number of the object.
        """
        return self._pyosmium_data.version()

    @property
    def changeset(self) -> int:
        """ (read-only) Id of changeset where this version of the object was created.
        """
        return self._pyosmium_data.changeset()

    @property
    def uid(self) -> int:
        """ (read-only) Id of the user that created this version
            of the object. Only this ID uniquely identifies users.
        """
        return self._pyosmium_data.uid()

    @property
    def timestamp(self) -> dt.datetime:
        """ (read-only) Date when this version has been created,
            returned as a ``datetime.datetime``.
        """
        return self._pyosmium_data.timestamp()

    @property
    def user(self) -> str:
        """ (read-only) Name of the user that created this version.
            Be aware that user names can change, so that the same
            user ID may appear with different names and vice versa.
        """
        return self._pyosmium_data.user()

    @property
    def tags(self) -> TagList:
        """ (read-only) List of tags describing the object.
            See [`osmium.osm.TagList`][].
        """
        return self._tags

    def positive_id(self) -> int:
        """ Get the absolute value of the id of this object.
        """
        return self._pyosmium_data.positive_id()

    def user_is_anonymous(self) -> bool:
        """ Check if the user is anonymous. If true, the uid does not uniquely
            identify a single user but only the group of all anonymous users
            in general.
        """
        return self._pyosmium_data.user_is_anonymous()

    def type_str(self) -> str:
        """ Return a single character identifying the type of the object.
            The character is the same as used in OPL.
        """
        return 'o'

    def is_node(self) -> bool:
        """ Return true if the object is a Node object.
        """
        return isinstance(self, Node)

    def is_way(self) -> bool:
        """ Return true if the object is a Way object.
        """
        return isinstance(self, Way)

    def is_relation(self) -> bool:
        """ Return true if the object is a Relation object.
        """
        return isinstance(self, Relation)

    def is_area(self) -> bool:
        """ Return true if the object is a Way object.
        """
        return isinstance(self, Area)


class Node(OSMObject['cosm.COSMNode']):
    """ Represents a single OSM node. It inherits all properties from OSMObjects
        and adds a single extra attribute: the location.
    """

    def __init__(self, cnode: 'cosm.COSMNode'):
        self._pyosmium_data = cnode
        self._location: Optional['osmium.osm.Location'] = None
        self._tags = TagList(self._pyosmium_data)

    def replace(self, **kwargs: Any) -> 'osmium.osm.mutable.Node':
        """ Create a mutable node replacing the properties given in the
            named parameters. The properties may be any of the properties
            of OSMObject or Node.

            Note that this function only creates a shallow copy per default.
            It is still bound to the scope of the original object.
            To create a full copy use:
            ``node.replace(tags=dict(node.tags))``
        """
        return osmium.osm.mutable.Node(self, **kwargs)

    @property
    def location(self) -> 'osmium.osm.Location':
        """ The geographic coordinates of the node.
            See [`osmium.osm.Location`][].
        """
        if self._location is None:
            self._location = self._pyosmium_data.location()

        return self._location

    @property
    def lat(self) -> float:
        """ Return latitude of the node.
        """
        return self.location.lat


    @property
    def lon(self) -> float:
        """ Return longitude of the node.
        """
        return self.location.lon

    def type_str(self) -> str:
        return 'n'

    def __str__(self) -> str:
        if self._pyosmium_data.is_valid():
            return f'n{self.id:d}: location={self.location!s} tags={self.tags!s}'

        return '<node invalid>'


    __repr__ = _make_repr('Node', 'id', 'deleted', 'visible', 'version',
                          'changeset', 'uid', 'timestamp', 'user',
                          'tags', 'location')


class Way(OSMObject['cosm.COSMWay']):
    """ Represents an OSM way. It inherits the attributes from OSMObject and
        adds an ordered list of nodes that describes the way.
    """

    def __init__(self, cway: 'cosm.COSMWay'):
        self._pyosmium_data = cway
        self._tags = TagList(self._pyosmium_data)
        self._nodes: Optional[WayNodeList] = None

    def replace(self, **kwargs: Any) -> 'osmium.osm.mutable.Way':
        """ Create a mutable way replacing the properties given in the
            named parameters. The properties may be any of the properties
            of OSMObject or Way.

            Note that this function only creates a shallow copy per default.
            It is still bound to the scope of the original object.
            To create a full copy use:
            ``way.replace(tags=dict(way.tags), nodes=list(way.nodes))``
        """
        return osmium.osm.mutable.Way(self, **kwargs)

    @property
    def nodes(self) -> WayNodeList:
        """ (read-only) Ordered list of nodes.
            See [`osmium.osm.WayNodeList`][].
        """
        if self._nodes is None:
            self._nodes = WayNodeList(self._pyosmium_data, self._pyosmium_data.nodes())

        return self._nodes

    def is_closed(self) -> bool:
        """ True if the start and end node are the same (synonym for
            ``ends_have_same_id``).
        """
        return self._pyosmium_data.is_closed()

    def ends_have_same_id(self) -> bool:
        """ True if the start and end node are exactly the same.
        """
        return self._pyosmium_data.is_closed()

    def ends_have_same_location(self) -> bool:
        """ True if the start and end node of the way are at the same location.
            Expects that the coordinates of the way nodes have been loaded
            (see [SimpleHandler apply functions][osmium.SimpleHandler] and
            [`FileProcessor.with_locations()`][osmium.FileProcessor.with_locations])
            If the locations are not present then the function returns always true.
        """
        return self._pyosmium_data.ends_have_same_location()

    def type_str(self) -> str:
        return 'w'

    def __str__(self) -> str:
        if self._pyosmium_data.is_valid():
            return f'w{self.id:d}: nodes={self.nodes!s} tags={self.tags!s}'

        return '<way invalid>'

    __repr__ = _make_repr('Way', 'id', 'deleted', 'visible', 'version', 'changeset',
                          'uid', 'timestamp', 'user', 'tags', 'nodes')


class Relation(OSMObject['cosm.COSMRelation']):
    """ Represents a OSM relation. It inherits the attributes from OSMObject
        and adds an ordered list of members.
    """
    _members: RelationMemberList

    def __init__(self, crelation: 'cosm.COSMRelation'):
        self._pyosmium_data = crelation
        self._tags = TagList(self._pyosmium_data)
        self._members = RelationMemberList(self._pyosmium_data)

    def replace(self, **kwargs: Any) -> 'osmium.osm.mutable.Relation':
        """ Create a mutable relation replacing the properties given in the
            named parameters. The properties may be any of the properties
            of OSMObject or Relation.

            Note that this function only creates a shallow copy per default.
            It is still bound to the scope of the original object.
            To create a full copy use:
            ``rel.replace(tags=dict(rel.tags), members=list(rel.members))``
        """
        return osmium.osm.mutable.Relation(self, **kwargs)

    @property
    def members(self) -> RelationMemberList:
        """(read-only) Ordered list of relation members.
           See [`osmium.osm.RelationMemberList`][].
        """
        return self._members

    def type_str(self) -> str:
        return 'r'

    def __str__(self) -> str:
        if self._pyosmium_data.is_valid():
            return f"r{self.id:d}: members={self.members!s}, tags={self.tags!s}"

        return f"<relation invalid>"

    __repr__ = _make_repr('Relation', 'id', 'deleted', 'visible', 'version',
                                      'changeset', 'uid', 'timestamp', 'user',
                                      'tags', 'members')


class OuterRingIterator:
    """ Iterator over outer rings of an area.
    """

    def __init__(self, parent: Any) -> None:
        self._pyosmium_data = parent
        self.iterator = self._pyosmium_data.outer_begin()

    def __iter__(self) -> 'OuterRingIterator':
        return self

    def __next__(self) -> OuterRing:
        return OuterRing(self._pyosmium_data, self._pyosmium_data.outer_next(self.iterator))


class InnerRingIterator:
    """ Iterator over the inner rings of an outer ring.
    """

    def __init__(self, parent: Any, oring: OuterRing) -> None:
        self._pyosmium_data = parent
        self.iterator = self._pyosmium_data.inner_begin(oring._list)

    def __iter__(self) -> 'InnerRingIterator':
        return self

    def __next__(self) -> InnerRing:
        return InnerRing(self._pyosmium_data, self._pyosmium_data.inner_next(self.iterator))


class Area(OSMObject['cosm.COSMArea']):
    """ Areas are a special kind of meta-object representing a polygon.
        They can either be derived from closed ways or from relations
        that represent multipolygons. They also inherit the attributes
        of OSMObjects and in addition contain polygon geometries. Areas have
        their own unique id space. This is computed as the OSM id times 2
        and for relations 1 is added.
    """

    def __init__(self, carea: 'cosm.COSMArea') -> None:
        self._pyosmium_data = carea
        self._tags = TagList(self._pyosmium_data)

    def from_way(self) -> bool:
        """ Return true if the area was created from a way, false if it was
            created from a relation of multipolygon type.
        """
        return self._pyosmium_data.from_way()

    def orig_id(self) -> int:
        """ Compute the original OSM id of this object. Note that this is not
            necessarily unique because the object might be a way or relation
            which have an overlapping id space.
        """
        return self._pyosmium_data.orig_id()

    def is_multipolygon(self) -> bool:
        """ Return true if this area is a true multipolygon, i.e. it consists
            of multiple outer rings.
        """
        return self._pyosmium_data.is_multipolygon()

    def num_rings(self) -> Tuple[int,int]:
        """ Return a tuple with the number of outer rings and inner rings.

            This function goes through all rings to count them.
        """
        return self._pyosmium_data.num_rings()

    def outer_rings(self) -> OuterRingIterator:
        """ Return an iterator over all outer rings of the multipolygon.
        """
        return OuterRingIterator(self._pyosmium_data)

    def inner_rings(self, oring: OuterRing) -> InnerRingIterator:
        """ Return an iterator over all inner rings of the multipolygon.
        """
        return InnerRingIterator(self._pyosmium_data, oring)

    def type_str(self) -> str:
        return 'a'

    def __str__(self) -> str:
        if self._pyosmium_data.is_valid():
            return f"a{self.id:d}: num_rings={self.num_rings()}, tags={self.tags!s}"

        return f"<area invalid>"

    __repr__ = _make_repr('Area', 'id', 'deleted', 'visible', 'version',
                                  'changeset', 'uid', 'timestamp', 'user',
                                  'tags')


class Changeset:
    """ A changeset description.
    """
    _pyosmium_data: 'cosm.COSMChangeset'
    _bounds: Optional['osmium.osm.Box']
    _tags: TagList

    def __init__(self, carea: 'cosm.COSMChangeset') -> None:
        self._pyosmium_data = carea
        self._bounds = None
        self._tags = TagList(self._pyosmium_data)

    @property
    def id(self) -> int:
        """ (read-only) Unique ID of the changeset.
        """
        return self._pyosmium_data.id()

    @property
    def uid(self) -> int:
        """ (read-only) User ID of the changeset creator.
        """
        return self._pyosmium_data.uid()

    @property
    def created_at(self) -> dt.datetime:
        """ (read-only) Timestamp when the changeset was first opened.
        """
        return self._pyosmium_data.created_at()

    @property
    def closed_at(self) -> dt.datetime:
        """ (read-only) Timestamp when the changeset was finalized. May be
            ``None`` when the changeset is still open.
        """
        return self._pyosmium_data.closed_at()

    @property
    def open(self) -> bool:
        """ (read-only) True when the changeset is still open.
        """
        return self._pyosmium_data.open()

    @property
    def num_changes(self) -> int:
        """ (read-only) The total number of objects changed in this Changeset.
        """
        return self._pyosmium_data.num_changes()

    @property
    def user(self) -> str:
        """ (read-only) Name of the user that created the changeset.
            Be aware that user names can change, so that the same
            user ID may appear with different names and vice versa.
        """
        return self._pyosmium_data.user()

    @property
    def bounds(self) -> 'osmium.osm.Box':
        """ (read-only) The bounding box of the area that was edited.
        """
        if self._bounds is None:
            self._bounds = self._pyosmium_data.bounds()

        return self._bounds

    @property
    def tags(self) -> TagList:
        """ (read-only) List of tags describing the object.
            See [`osmium.osm.TagList`][].
        """
        return self._tags

    def user_is_anonymous(self) -> bool:
        """ Check if the user anonymous. If true, the uid does not uniquely
            identify a single user but only the group of all anonymous users
            in general.
        """
        return self._pyosmium_data.user_is_anonymous()

    def type_str(self) -> str:
        return 'c'

    def __str__(self) -> str:
        if self._pyosmium_data.is_valid():
            return f'c{self.id:d}: closed_at={self.closed_at!s}, bounds={self.bounds!s}, tags={self.tags!s}'

        return f"<changeset invalid>"

    __repr__ =  _make_repr('Changeset', 'id', 'uid', 'created_at', 'closed_at',
                                        'open', 'num_changes', 'bounds', 'user',
                                        'tags')


OSMEntity = Union[Node, Way, Relation, Area, Changeset]



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2023 Sarah Hoffmann.
""" Tests for the pyosmium-get-changes script.
"""
from pathlib import Path
from textwrap import dedent

import pytest
import osmium.replication.server
import osmium as o

from helpers import IDCollector

try:
    import http.cookiejar as cookiejarlib
except ImportError:
    import cookielib as cookiejarlib


class TestPyosmiumGetChanges:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.script = dict()

        filename = (Path(__file__) / ".." / ".." / "tools"/ "pyosmium-get-changes").resolve()
        with filename.open("rb") as f:
            exec(compile(f.read(), str(filename), 'exec'), self.script)


    def main(self, httpserver, *args):
        return self.script['main'](['--server', httpserver.url_for('')] + list(args))


    def test_init_id(self, capsys, httpserver):
        assert 0 == self.main(httpserver, '-I', '453')

        output = capsys.readouterr().out.strip()

        assert output == '453'


    def test_init_date(self, capsys, httpserver):
        httpserver.expect_request('/state.txt').respond_with_data(dedent("""\
                    sequenceNumber=100
                    timestamp=2017-08-26T11\\:04\\:02Z
                 """))
        httpserver.expect_request('/000/000/000.state.txt').respond_with_data(dedent("""\
                    sequenceNumber=0
                    timestamp=2016-08-26T11\\:04\\:02Z
                 """))
        assert 0 == self.main(httpserver, '-D', '2015-12-24T08:08:08Z')

        output = capsys.readouterr().out.strip()

        assert output == '-1'


    def test_init_to_file(self, tmp_path, httpserver):
        fname = tmp_path / 'db.seq'

        assert 0 == self.main(httpserver, '-I', '453', '-f', str(fname))
        assert fname.read_text() == '453'


    def test_init_from_seq_file(self, tmp_path, httpserver):
        fname = tmp_path / 'db.seq'
        fname.write_text('453')

        assert 0 == self.main(httpserver, '-f', str(fname))
        assert fname.read_text() == '453'


    def test_init_date_with_cookie(self, capsys, tmp_path, httpserver):
        httpserver.expect_request('/state.txt').respond_with_data(dedent("""\
                    sequenceNumber=100
                    timestamp=2017-08-26T11\\:04\\:02Z
                 """))
        httpserver.expect_request('/000/000/000.state.txt').respond_with_data(dedent("""\
                    sequenceNumber=0
                    timestamp=2016-08-26T11\\:04\\:02Z
                 """))

        fname = tmp_path / 'my.cookie'
        cookie_jar = cookiejarlib.MozillaCookieJar(str(fname))
        cookie_jar.save()

        assert 0 == self.main(httpserver, '--cookie', str(fname),
                              '-D', '2015-12-24T08:08:08Z')

        output = capsys.readouterr().out.strip()

        assert output == '-1'


    def test_get_simple_update(self, tmp_path, httpserver):
        outfile = tmp_path / 'outfile.opl'

        httpserver.expect_request('/state.txt').respond_with_data(dedent("""\
                    sequenceNumber=454
                    timestamp=2017-08-26T11\\:04\\:02Z
                 """))
        httpserver.expect_request('/000/000/454.state.txt').respond_with_data(dedent("""\
                    sequenceNumber=454
                    timestamp=2016-08-26T11\\:04\\:02Z
                 """))
        httpserver.expect_request('/000/000/454.opl').respond_with_data(
                 "n12 v1 x4 y6\nn13 v1 x9 y-6\nw2 v2 Nn1,n2")

        assert 0 == self.main(httpserver, '--diff-type', 'opl',
                              '-I', '453', '-o', str(outfile))

        ids = IDCollector()
        o.apply(str(outfile), ids)

        assert ids.nodes == [12, 13]
        assert ids.ways == [2]
        assert ids.relations == []



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import ByteString, Union, Optional, Any
import os

from .osm import osm_entity_bits
from .osm.types import OSMEntity
from .index import LocationTable, IdSet
from .io import Reader, Writer, Header, File, FileBuffer

# Placeholder for more narrow type definition to come
HandlerLike = object

class InvalidLocationError(Exception):
    """ Raised when the location of a node is requested that has
        no valid location. To be valid, a location must be
        inside the -180 to 180 and -90 to 90 degree range.
    """


class BaseHandler:
    """ Base class for all native handler functions in pyosmium.
        Any class that derives from this class can be used for
        parameters that need a handler-like object.
    """


class BaseFilter(BaseHandler):
    """ Base class for all native filter functions in pyosmium.
        A filter is a handler that returns a boolean in the handler
        functions indicating if the object should pass the filter (False)
        or be dropped (True).
    """
    def enable_for(self, entities: osm_entity_bits) -> None:
        """ Set the OSM types this filter should be applied to. If
            an object has a type for which the filter is not enabled,
            the filter will be skipped completely. Or to put it in
            different words: every object for which the filter is not
            enabled, passes the filter automatically.
        """


class BufferIterator:
    """ (internal) Iterator interface for reading from a queue of buffers.

        This class is needed for pyosmium's internal implementation. There is
        currently no way to create buffers or add them to the iterator
        from Python.
    """
    def __init__(self, *handlers: HandlerLike) -> None:
        """ Create a new iterator. The iterator will pass each
            object through the filter chain _handlers_ before returning
            it.
        """
    def __bool__(self) -> bool:
        """ True if there are any objects left to return.
        """
    def __iter__(self) -> 'BufferIterator':
        """ Returns itself.
        """
    def __next__(self) -> OSMEntity:
        """ Get the next OSM object from the buffer or raise an StopIteration.
        """


class MergeInputReader:
    """ Buffer which collects data from multiple input files, sorts it
        and optionally deduplicates the data before applying to a handler.
    """
    def __init__(self) -> None:
        """ Initialize a new reader.
        """
    def _apply_internal(self, *handlers: HandlerLike, simplify: bool = ...) -> None:
        """ Internal application function. Do not use.
        """
    def add_buffer(self, buffer: Union[ByteString, str], format: str) -> int:
        """ Add input data from a buffer to the reader. The buffer may
            be any data which follows the Python buffer protocol. The
            mandatory _format_ parameter describes the format of the data.

            The data will be copied into internal buffers, so that the input
            buffer can be safely discarded after the function has been called.
        """
    def add_file(self, file: str) -> int:
        """ Add data from the given input file _file_ to the reader.
        """
    def apply_to_reader(self, reader: Reader, writer: Writer, with_history: bool = ...) -> None:
        """ Apply the collected data to data from the given _reader_ and write
            the result to _writer_. This function can be used to merge the diff "
            data together with other OSM data (for example when updating a
            planet file. If _with_history_ is true, then the collected data will
            be applied verbatim without removing duplicates. This is important
            when using OSM history files as input.
        """
    def apply(self, *handlers: HandlerLike, idx: str = '', simplify: bool = True) -> None:
        """ Apply collected data to a handler. The data will be sorted first.
            If _simplify_ is true (default) then duplicates will be eliminated
            and only the newest version of each object kept. If _idx_ is given
            a node location cache with the given type will be created and
            applied when creating the ways. Note that a diff file normally does
            not contain all node locations to reconstruct changed ways. If the
            full way geometries are needed, create a persistent node location
            cache during initial import of the area and reuse it when processing
            diffs. After the data
            has been applied the buffer of the MergeInputReader is empty and
            new data can be added for the next round of application.
        """


class SimpleWriter(BaseHandler):
    """ Basic writer for OSM data. The SimpleWriter can write out
        object that are explicitly passed or function as a handler and
        write out all objects it receives. It is also possible to
        mix these two modes of operations.

        The writer writes out the objects in the order it receives them.
        It is the responsibility of the caller to ensure to follow the
        [ordering conventions][order-in-osm-files]
        for OSM files.

        The SimpleWriter should normally used as a context manager. If you
        don't use it in a `with` context, don't forget to call `close()`,
        when writing is finished.
    """
    def __init__(self, file: Union[str, 'os.PathLike[str]', File],
                 bufsz: int= ...,
                 header: Optional[Header]= ..., overwrite: bool= ...,
                 filetype: str= ...) -> None:
        """ Initiate a new writer for the given file. The writer will
            refuse to overwrite an already existing file unless _overwrite_
            is explicitly set to `True`.

            The file type is usually determined from the file extension.
            If you want to explicitly set the filetype (for example, when
            writing to standard output '-'), then use a File object.
            Using the _filetype_ parameter to set the file type is deprecated
            and only works when the file is a string.

            The _header_ parameter can be used to set a custom header in
            the output file. What kind of information can be written into
            the file header depends on the file type.

            The optional parameter _bufsz_ sets the size of the buffers used
            for collecting the data before they are written out. The default
            size is 4MB. Larger buffers are normally better but you should
            be aware that there are normally multiple buffers in use during
            the write process.
        """
    def add_node(self, node: object) -> None:
        """ Add a new node to the file. The node may be a
            [Node][osmium.osm.Node] object or its mutable
            variant or any other Python object that implements the same
            attributes.
        """
    def add_relation(self, relation: object) -> None:
        """ Add a new relation to the file. The relation may be a
            [Relation][osmium.osm.Relation] object or its mutable
            variant or any other Python object that implements the same
            attributes.
        """
    def add_way(self, way: object) -> None:
        """ Add a new way to the file. The way may be a
            [Way][osmium.osm.Way] object or its mutable
            variant or any other Python object that implements the same
            attributes.
        """
    def add(self, obj: object) -> None:
        """ Add a new object to the file. The function will try to determine
            the kind of object automatically.
        """
    def close(self) -> None:
        """ Flush the remaining buffers and close the writer. While it is not
            strictly necessary to call this function explicitly, it is still
            strongly recommended to close the writer as soon as possible, so
            that the buffer memory can be freed.
        """
    def __enter__(self) -> 'SimpleWriter':...
    def __exit__(self, *args: Any) -> None:...


class NodeLocationsForWays(BaseHandler):
    """ Handler for retriving and caching locations from ways
        and adding them to ways.
    """
    @property
    def apply_nodes_to_ways(self) -> bool:
        """ When set (the default), the collected locations
            are propagated to the node list of ways.
        """
    @apply_nodes_to_ways.setter
    def apply_nodes_to_ways(self, value: bool) -> None:...

    def __init__(self, locations: LocationTable) -> None:
        """ Initiate a new handler using the given location table _locations_
            to cache the node coordinates.
        """
    def ignore_errors(self) -> None:
        """ Disable raising an exception when filling the node list of
            a way and a coordinate is not available.
        """


class OsmFileIterator:
    """ Low-level iterator interface for reading from an OSM source.
    """
    def __init__(self, reader: Reader, *handlers: HandlerLike) -> None:
        """ Initialise a new iterator using the given _reader_ as source.
            Each object is passed through the list of filters given by
            _handlers_. If all the filters are passed, the object is
            returned by `next()`.
        """
    def set_filtered_handler(self, handler: object) -> None:
        """ Set a fallback handler for objects that have been filtered
            out. The objects will be passed to the single handler.
        """
    def __iter__(self) -> 'OsmFileIterator':
        """ Returns itself.
        """
    def __next__(self) -> OSMEntity:
        """ Get the next OSM object from the file or raise a StopIteration.
        """


class IdTrackerIdFilter(BaseFilter): ...


class IdTrackerContainsFilter(BaseFilter): ...


class IdTracker:
    """ Class to keep track of node, way and relation IDs.

        Ids can be added to the to the tracker in various ways: by
        adding IDs directly, by adding the IDs of referenced IDs in an
        OSM object or by extracting the referenced IDs from an input file.

        The tracker can then be used as a filter to select objects based
        on whether they are contained in the tracker's ID lists.
    """
    def __init__(self) -> None:
        """ Initialise a new empty tracker.
        """
    def add_node(self, node: int) -> None:
        """ Add the given node ID to the tracker.
        """
    def add_relation(self, relation: int) -> None:
        """ Add the given relation ID to the tracker.
        """
    def add_way(self, way: int) -> None:
        """ Add the given way ID to the tracker.
        """
    def add_references(self, obj: object) -> None:
        """ Add all IDs referenced by the input object _obj_.

            The function will track the IDs of node lists from
            [Way][osmium.osm.Way] objects or Python
            objects with a `nodes` attribute, which must be a sequence of ints.
            It also tracks the IDs of relation members from
            [Relation][osmium.osm.Relation] objects
            or Python objects with a `members` attribute with an
            equivalent content. Input objects that do not fall into
            any of these categories are silently ignored.
        """
    def contains_any_references(self, obj: object) -> bool:
        """ Check if the given input object _obj_ contains any references
            to IDs tracked by this tracker.

            The function will check the IDs of node lists from
            [Way][osmium.osm.Way] objects or Python
            objects with a `nodes` attribute, which must be a sequence of ints.
            It also tracks the IDs of relation members from
            [Relation][osmium.osm.Relation] objects
            or Python objects with a `members` attribute with an
            equivalent content. All other object kinds will return
            `False`.
        """
    def complete_backward_references(self, filename: Union[str, 'os.PathLike[str]', File, FileBuffer],
                                     relation_depth: int = ...) -> None:
        """ Make the IDs in the tracker reference-complete by adding
            all referenced IDs for objects whose IDs are already tracked.

            The function scans through the reference file `filename`, finds
            all the objects this tracker references and applies `add_references()`
            to them. The reference file is expected to be
            [sorted][order-in-osm-files].

            The _relation_depth_ parameter controls how nested relations are
            handled. When set to 0 then only way and node references of
            relations that are already tracked are completed. If the parameter
            is larger than 0, the function will make at a maximum
            _relation_depth_ passes through the reference file, to find
            nested relation. That means, that nested relations with a nesting
            depth up to _relation_depth_ are guaranteed to be included.
            Relations that are nested more deeply, may or may not appear.
        """
    def complete_forward_references(self, filename: Union[str, 'os.PathLike[str]', File, FileBuffer],
                                    relation_depth: int = ...) -> None:
        """ Add to the tracker all IDs of object that reference any ID already
            tracked.

            The function scans through the reference file `filename`,
            checks all objects in the file with the `contains_any_references()`
            function and adds the object ID to the tracker if the check
            is positive.

            The _relation_depth_ parameter controls how nested relations are
            handled. When set to a value smaller than 0, then relations will
            no be added at all to the tracker. When set to 0, then only
            relations are added that reference a node or way already in the
            tracker. When set to a strictly positive value, then nested
            relations are tacken into account as well. The function will
            make at a maximum _relation_depth_ passes to complete relations
            with relation members.
        """
    def id_filter(self) -> IdTrackerIdFilter:
        """ Return a filter object which lets all nodes, ways and relations
            pass that are being tracked in this tracker.

            You may change the tracker while the filter is in use. Such
            a change is then immediately reflected in the filter.

            The filter has no effect on areas and changesets.
        """
    def contains_filter(self) -> IdTrackerContainsFilter:
        """ Return a filter object that lets all ways and relations pass
            which reference any of the object IDs tracked by this tracker.

            You may change the tracker while the filter is in use. Such
            a change is then immediately reflected in the filter.

            The filter has no effect on nodes, areas and changesets.
        """
    def node_ids(self) -> IdSet:
        """ Return a view of the set of node ids. The returned object is
            mutable. You may call operations like `unset()` and `clear()`
            on it, which then have a direct effect on the tracker.
        """
    def way_ids(self) -> IdSet:
        """ Return a view of the set of way ids.The returned object is
            mutable. You may call operations like `unset()` and `clear()`
            on it, which then have a direct effect on the tracker.
        """
    def relation_ids(self) -> IdSet:
        """ Return a view of the set of relation ids. The returned object is
            mutable. You may call operations like `unset()` and `clear()`
            on it, which then have a direct effect on the tracker.
        """

def apply(reader: Union[Reader, str, 'os.PathLike[str]', File, FileBuffer],
          *handlers: HandlerLike) -> None:
    """ Apply a chain of handlers to the given input source. The input
        source may be a [osmium.io.Reader][], a file or a file buffer.
        If one of the handlers is a
        [filter][osmium.BaseFilter], then processing of the
        object will be stopped when it does not pass the filter.
    """



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest

import osmium as o

from helpers import IDCollector

@pytest.fixture
def ref_file(test_data):
    return test_data("""\
                n1 x0 y0
                n2 x0 y0
                n99 x0 y0
                n100 x0 y0
                w12 Nn1,n2
                w90 Nn10,n11
                r2 Mn99@
                r10 Mn100@,w90@,r2@
                """)


class DummyNode:
    def __init__(self, nid):
        self.id = nid
        self.location = (3, 4)


def test_simple_forward_no_back_reference(ref_file, tmp_path):
    outfile = str(tmp_path / 'test.osm')

    with o.ForwardReferenceWriter(outfile, ref_file, back_references=False) as writer:
        writer.add_node(DummyNode(2))
        writer.add_node(DummyNode(99))

    ids = IDCollector()
    o.apply(outfile, ids)

    assert ids.nodes == [2, 99]
    assert ids.ways == [12]
    assert ids.relations == [2]


def test_simple_forward_with_back_reference(ref_file, tmp_path):
    outfile = str(tmp_path / 'test.osm')

    with o.ForwardReferenceWriter(outfile, ref_file) as writer:
        writer.add_node(DummyNode(2))
        writer.add_node(DummyNode(99))

    ids = IDCollector()

    for obj in o.FileProcessor(outfile)\
                .with_filter(ids)\
                .with_filter(o.filter.EntityFilter(o.osm.NODE)):
        if obj.id in (2, 99):
            assert obj.lat == 4
            assert obj.lon == 3
        else:
            assert obj.lat == 0
            assert obj.lon == 0

    assert ids.nodes == [1, 2, 99]
    assert ids.ways == [12]
    assert ids.relations == [2]



"""
Converts a file from one format to another.

This example shows how to write objects to a file.
"""

import osmium as o

import sys

if __name__ == '__main__':
    if len(sys.argv) != 3:
        print("Usage: python convert.py <infile> <outfile>")
        sys.exit(-1)

    writer = o.SimpleWriter(sys.argv[2])

    for obj in o.FileProcessor(sys.argv[1]):
        if obj.is_node():
            writer.add_node(obj)
        elif obj.is_way():
            writer.add_way(obj)
        elif obj.is_relation():
            writer.add_relation(obj)

    writer.close()




# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2023 Sarah Hoffmann.
import logging
import time
from textwrap import dedent

import pytest

from werkzeug.wrappers import Response

from helpers import mkdate, CountingHandler

import osmium as o
import osmium.replication.server as rserv
import osmium.replication


@pytest.mark.parametrize("inp,outp", [
        (None,      'https://text.org/state.txt'),
        (1,         'https://text.org/000/000/001.state.txt'),
        (999,       'https://text.org/000/000/999.state.txt'),
        (1000,      'https://text.org/000/001/000.state.txt'),
        (573923,    'https://text.org/000/573/923.state.txt'),
        (3290012,   'https://text.org/003/290/012.state.txt'),
    ])
def test_get_state_url(inp, outp):
    svr = rserv.ReplicationServer("https://text.org")

    assert outp == svr.get_state_url(inp)


@pytest.mark.parametrize("inp,outp", [
        (1,         'https://who.is/me//000/000/001.osc.gz'),
        (500,       'https://who.is/me//000/000/500.osc.gz'),
        (83750,     'https://who.is/me//000/083/750.osc.gz'),
        (999999999, 'https://who.is/me//999/999/999.osc.gz'),
    ])
def test_get_diff_url(inp, outp):
    svr = rserv.ReplicationServer("https://who.is/me/")

    assert outp, svr.get_diff_url(inp)


def test_get_newest_change_from_file(tmp_path):
    fn = tmp_path / 'change.opl'
    fn.write_text('n6365 v1 c63965061 t2018-10-29T03:56:07Z i8369524 ux x1 y7')

    val = osmium.replication.newest_change_from_file(str(fn))
    assert val == mkdate(2018, 10, 29, 3, 56, 7)


def test_get_state_valid(httpserver):
    httpserver.expect_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z
        txnReadyList=
        txnMax=1219304113
        txnActiveList=1219303583,1219304054,1219304104""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_sequence_cut(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=259""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_date_cut(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-2""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_timestamp_cut(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_permanent_error(httpserver, caplog):
    httpserver.expect_request('/state.txt').respond_with_data('stuff', status=404)

    with caplog.at_level(logging.DEBUG):
        res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is None
    assert "Loading state info failed" in caplog.text


def test_get_state_transient_error(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data('stuff', status=500)
    httpserver.expect_ordered_request('/state.txt').respond_with_data('stuff', status=500)
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z
        txnReadyList=
        txnMax=1219304113
        txnActiveList=1219303583,1219304054,1219304104""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_too_many_retries(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is None


def test_get_state_server_timeout(httpserver):
    def sleeping(request):
        time.sleep(2)
        return Response("""\
    #Sat Aug 26 11:04:04 UTC 2017
    txnMaxQueried=1219304113
    sequenceNumber=2594669
    timestamp=2017-08-26T11\\:04\\:02Z
    txnReadyList=
    txnMax=1219304113
    txnActiveList=1219303583,1219304054,1219304104""")

    httpserver.expect_request("/state.txt").respond_with_handler(sleeping)

    with rserv.ReplicationServer(httpserver.url_for('')) as svr:
        svr.set_request_parameter('timeout', 1)

        res = svr.get_state_info()

    assert res is None


def test_apply_diffs_count(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        assert 100 == svr.apply_diffs(h, 100, 10000)
        assert h.counts == [1, 1, 1, 0]


def test_apply_diffs_without_simplify(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 v23
        n1 v24
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        assert 100 == svr.apply_diffs(h, 100, 10000, simplify=False)
        assert [2, 1, 1, 0] == h.counts


def test_apply_diffs_with_simplify(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 v23
        n1 v24
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        assert 100 == svr.apply_diffs(h, 100, 10000, simplify=True)
        assert [1, 1, 1, 0] == h.counts


def test_apply_with_location(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))

    class Handler(CountingHandler):
        def way(self, w):
            self.counts[1] += 1
            assert 2 == len(w.nodes)
            assert 1 == w.nodes[0].ref
            assert 10 == w.nodes[0].location.lon
            assert 23 == w.nodes[0].location.lat
            assert 2 == w.nodes[1].ref
            assert not w.nodes[1].location.valid()

    h = Handler()
    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        assert 100 == svr.apply_diffs(h, 100, 10000, idx="flex_mem")
        assert h.counts == [1, 1, 0, 0]


def test_apply_reader_without_simplify(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 v23
        n1 v24
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        diffs = svr.collect_diffs(100, 100000)

    assert diffs is not None
    diffs.reader.apply(h, simplify=False)
    assert [2, 1, 1, 0] == h.counts


def test_apply_reader_with_simplify(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 v23
        n1 v24
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        diffs = svr.collect_diffs(100, 100000)

    assert diffs is not None
    diffs.reader.apply(h, simplify=True)
    assert [1, 1, 1, 0] == h.counts


def test_apply_reader_with_location(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))

    class Handler(CountingHandler):
        def way(self, w):
            self.counts[1] += 1
            assert 2 == len(w.nodes)
            assert 1 == w.nodes[0].ref
            assert 10 == w.nodes[0].location.lon
            assert 23 == w.nodes[0].location.lat
            assert 2 == w.nodes[1].ref
            assert not w.nodes[1].location.valid()

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = Handler()
        diffs = svr.collect_diffs(100, 100000)

    assert diffs is not None
    diffs.reader.apply(h, idx="flex_mem")
    assert h.counts == [1, 1, 0, 0]


def test_apply_diffs_permanent_error(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl')\
              .respond_with_data('not a file', status=404)

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            assert None == svr.apply_diffs(h, 100, 10000)
            assert h.counts == [0, 0, 0, 0]

    assert 'Error during diff download' in caplog.text


def test_apply_diffs_permanent_error_later_diff(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=101
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))
    httpserver.expect_ordered_request('/000/000/101.opl')\
              .respond_with_data('not a file', status=404)

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            assert 100 == svr.apply_diffs(h, 100, 10000)
            assert h.counts == [1, 1, 0, 0]

    assert 'Error during diff download' in caplog.text


def test_apply_diffs_transient_error(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=101
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))
    httpserver.expect_ordered_request('/000/000/101.opl')\
              .respond_with_data('not a file', status=503)
    httpserver.expect_ordered_request('/000/000/101.opl').respond_with_data(dedent("""\
        n2 x10.0 y23.0
    """))

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            assert 101 == svr.apply_diffs(h, 100, 10000)
            assert h.counts == [2, 1, 0, 0]

    assert 'Error during diff download' not in caplog.text




def test_apply_diffs_transient_error_permanent(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=101
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))
    for _ in range(4):
        httpserver.expect_ordered_request('/000/000/101.opl')\
                  .respond_with_data('not a file', status=503)
    httpserver.expect_ordered_request('/000/000/101.opl').respond_with_data(dedent("""\
        n2 x10.0 y23.0
    """))

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            assert 100 == svr.apply_diffs(h, 100, 10000)
            assert h.counts == [1, 1, 0, 0]

    assert 'Error during diff download' in caplog.text



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2024 Sarah Hoffmann.
import re
from itertools import count
import pytest

from helpers import mkdate

import osmium as o

def apply_simple(handler, data, locations, tmp_path):
    handler.apply_buffer(data.encode('utf-8'), 'opl', locations=locations)


def apply_with_merge(handler, data, locations, tmp_path):
    mir = o.MergeInputReader()

    mir.add_buffer(data.encode('utf-8'), format='opl')

    fn = tmp_path / 'temp.opl'
    fn.write_text(data)
    mir.add_file(str(fn))

    mir.apply(handler, idx='flex_mem' if locations else '')


def _make_importer_factory(apply_func, tmp_path, to_opl):

    def _run(data, node=None, way=None, relation=None, area=None, changeset=None, locations=False):
        cnt = count()
        def _m(func):
            if func is None:
                return None

            def _inner(obj):
                next(cnt)
                func(obj)
            return _inner

        handler = o.make_simple_handler(node=_m(node), way=_m(way),
                                        relation=_m(relation), area=_m(area),
                                        changeset=_m(changeset))
        apply_func(handler, to_opl(data), locations, tmp_path)

        return next(cnt)

    return _run


@pytest.fixture(params=[apply_simple, apply_with_merge])
def test_importer(request, tmp_path, to_opl):
    return _make_importer_factory(request.param, tmp_path, to_opl)


@pytest.fixture(params=[apply_simple])
def area_importer(request, tmp_path, to_opl):
    return _make_importer_factory(request.param, tmp_path, to_opl)


def test_invalid_location():
    loc = o.osm.Location()
    assert not loc.valid()
    assert str(loc) == 'invalid'
    assert repr(loc) == 'osmium.osm.Location()'

    with pytest.raises(o.InvalidLocationError):
        lat = loc.lat
    with pytest.raises(o.InvalidLocationError):
        lon = loc.lon
    # these two don't raise an exception
    assert loc.lat_without_check() is not None
    assert loc.lon_without_check() is not None


def test_valid_location():
    loc = o.osm.Location(-1, 10)
    assert loc.lon == pytest.approx(-1)
    assert loc.lat == pytest.approx(10)
    assert loc.x == -10000000
    assert loc.y == 100000000
    assert re.fullmatch('-1.0*/10.0*', str(loc))
    assert repr(loc) == 'osmium.osm.Location(x=-10000000, y=100000000)'


@pytest.mark.parametrize('attrname', ['id', 'deleted', 'visible',
                                      'changeset', 'uid', 'timestamp', 'user',
                                      'tags'])
@pytest.mark.parametrize('osmdata', ['n1 v5 c58674 uänonymous',
                                     'w34 Nn34',
                                     'r45 Tfoo=rte'])
def test_object_attribute_do_not_overwrite(opl_buffer, attrname, osmdata):
    for n in o.FileProcessor(opl_buffer(osmdata)):
        with pytest.raises(AttributeError):
            setattr(n, attrname, 3)


def test_node_attributes(opl_buffer):
    node_data = 'n1 v5 c58674 t2014-01-31T06:23:35Z i42 uänonymous'

    for n in o.FileProcessor(opl_buffer(node_data)):
        assert n.deleted == False
        assert n.visible == True
        assert n.version == 5
        assert n.changeset == 58674
        assert n.uid == 42
        assert n.user_is_anonymous() == False
        assert n.timestamp == mkdate(2014, 1, 31, 6, 23, 35)
        assert n.user == u'änonymous'
        assert n.positive_id() == 1
        assert n.is_node()
        assert not n.is_way()
        assert n.type_str() == 'n'
        assert str(n) == 'n1: location=invalid tags={}'
        assert repr(n) == "osmium.osm.Node(id=1, deleted=False, visible=True, version=5, changeset=58674, uid=42, timestamp=datetime.datetime(2014, 1, 31, 6, 23, 35, tzinfo=datetime.timezone.utc), user='änonymous', tags=osmium.osm.TagList({}), location=osmium.osm.Location())"
        break
    else:
        assert False


def test_node_location(opl_buffer):
    for n in o.FileProcessor(opl_buffer("n1 x4 y5")):
        assert n.lat == 5.0
        assert n.lon == 4.0


@pytest.mark.parametrize('nid', (23, 0, -68373, 17179869418, -17179869417))
def test_node_positive_id(opl_buffer, nid):
    for n in o.FileProcessor(opl_buffer(f"n{nid} v5 c58674")):
        assert n.id == nid
        assert n.positive_id() == abs(nid)
        break
    else:
        assert False


def test_way_attributes(test_importer):
    def way(o):
        assert o.id == 1
        assert o.deleted == False
        assert o.visible == True
        assert o.version == 5
        assert o.changeset == 58674
        assert o.uid == 42
        assert o.user_is_anonymous() == False
        assert o.timestamp == mkdate(2014, 1, 31, 6, 23, 35)
        assert o.user == 'anonymous'
        assert o.positive_id() == 1
        assert o.is_way()
        assert o.type_str() == 'w'
        assert not o.is_closed()
        assert not o.ends_have_same_id()
        assert not o.ends_have_same_location()

        assert str(o) == 'w1: nodes=[1@0.0000000/0.0000000,2,3@1.0000000/1.0000000] tags={}'
        assert repr(o) == "osmium.osm.Way(id=1, deleted=False, visible=True, version=5, changeset=58674, uid=42, timestamp=datetime.datetime(2014, 1, 31, 6, 23, 35, tzinfo=datetime.timezone.utc), user='anonymous', tags=osmium.osm.TagList({}), nodes=osmium.osm.WayNodeList([osmium.osm.NodeRef(ref=1, location=osmium.osm.Location(x=0, y=0)), osmium.osm.NodeRef(ref=2, location=osmium.osm.Location()), osmium.osm.NodeRef(ref=3, location=osmium.osm.Location(x=10000000, y=10000000))]))"

        assert str(o.nodes) == '[1@0.0000000/0.0000000,2,3@1.0000000/1.0000000]'
        assert repr(o.nodes) == "osmium.osm.WayNodeList([osmium.osm.NodeRef(ref=1, location=osmium.osm.Location(x=0, y=0)), osmium.osm.NodeRef(ref=2, location=osmium.osm.Location()), osmium.osm.NodeRef(ref=3, location=osmium.osm.Location(x=10000000, y=10000000))])"

    assert 1 == test_importer(['n1 x0 y0', 'n3 x1 y1',
                               'w1 v5 c58674 t2014-01-31T06:23:35Z i42 uanonymous Nn1,n2,n3'],
                              way=way, locations=True)


def test_way_attribute_do_not_overwrite(opl_buffer):
    data = """\
           w34 Nn34
           """
    for w in o.FileProcessor(opl_buffer(data)):
        with pytest.raises(AttributeError):
            w.nodes = [3,4,5]


def test_relation_attributes(test_importer):
    def relation(o):
        assert o.id == 1
        assert o.deleted == False
        assert o.visible == True
        assert o.version == 5
        assert o.changeset == 58674
        assert o.uid == 42
        assert o.user_is_anonymous() == False
        assert o.timestamp == mkdate(2014, 1, 31, 6, 23, 35)
        assert o.user == ' anonymous'
        assert o.positive_id() == 1
        assert o.is_relation()
        assert o.type_str() == 'r'

        assert str(o) == 'r1: members=[w1], tags={}'
        assert repr(o) == "osmium.osm.Relation(id=1, deleted=False, visible=True, version=5, changeset=58674, uid=42, timestamp=datetime.datetime(2014, 1, 31, 6, 23, 35, tzinfo=datetime.timezone.utc), user=' anonymous', tags=osmium.osm.TagList({}), members=osmium.osm.RelationMemberList([osmium.osm.RelationMember(ref=1, type='w', role='')]))"

        assert str(o.members) == '[w1]'
        assert repr(o.members) == "osmium.osm.RelationMemberList([osmium.osm.RelationMember(ref=1, type='w', role='')])"

    assert 1 == test_importer('r1 v5 c58674 t2014-01-31T06:23:35Z i42 u%20%anonymous Mw1@',
                              relation=relation)


def test_relation_attribute_do_not_overwrite(opl_buffer):
    data = """\
           r34 Mn23@,w34@
           """
    for r in o.FileProcessor(opl_buffer(data)):
        with pytest.raises(AttributeError):
            r.members = [3,4,5]


def test_area_from_way_attributes(area_importer):
    def area(o):
        assert o.id == 46
        assert o.deleted == False
        assert o.visible == True
        assert o.version == 5
        assert o.changeset == 58674
        assert o.uid == 42
        assert o.user_is_anonymous() == False
        assert o.timestamp == mkdate(2014, 1, 31, 6, 23, 35)
        assert o.user == 'anonymous'
        assert o.positive_id() == 46
        assert o.orig_id() == 23
        assert o.from_way() == True
        assert o.is_area()
        assert o.type_str() == 'a'
        assert o.is_multipolygon() == False
        assert o.num_rings() == (1, 0)
        assert len(list(o.outer_rings())) == 1
        oring = list(o.outer_rings())[0]
        assert len(list(oring)) == 4
        assert set((1,2,3)) == set([x.ref for x in oring])
        assert oring.is_closed()
        assert oring.ends_have_same_id()
        assert oring.ends_have_same_location()
        assert len(list(o.inner_rings(oring))) == 0

    assert 1 == area_importer(['n1 x0 y0', 'n2 x1 y0', 'n3 x0 y1',
                               'w23 v5 c58674 t2014-01-31T06:23:35Z i42 uanonymous Nn1,n2,n3,n1 Tarea=yes'],
                              area=area)


@pytest.mark.parametrize('mptype', ('multipolygon', 'boundary'))
def test_area_from_multipolygon_relation(area_importer, mptype):
    def area(o):
        assert o.id == 3
        assert o.deleted == False
        assert o.visible == True
        assert o.version == 3
        assert o.changeset == 7654
        assert o.uid == 42
        assert o.user_is_anonymous() == False
        assert o.timestamp == mkdate(2014, 1, 31, 6, 23, 35)
        assert o.user == 'Anon'
        assert o.positive_id() == 3
        assert o.orig_id() == 1
        assert o.from_way() == False
        assert o.is_multipolygon() == False
        assert o.num_rings() == (1, 0)
        assert len(list(o.outer_rings())) == 1
        oring = list(o.outer_rings())[0]
        assert len(list(oring)) == 4
        assert set((1,2,3)) == set([x.ref for x in oring])
        assert oring.is_closed()
        assert oring.ends_have_same_id()
        assert oring.ends_have_same_location()
        assert len(list(o.inner_rings(oring))) == 0

    assert 1 == area_importer(['n1 x0 y0', 'n2 x1 y0', 'n3 x0 y1',
                               'w23 Nn1,n2,n3', 'w24 Nn3,n1',
                               'r1 v3 c7654 t2014-01-31T06:23:35Z i42 uAnon Mw23@outer,w24@outer Ttype={}'.format(mptype)],
                               area=area)



def test_changest_attributes(area_importer):
    def changeset(c):
        assert 34 == c.id
        assert 1 == c.uid
        assert not c.user_is_anonymous()
        assert "Steve" == c.user
        assert mkdate(2005, 4, 9, 19, 54, 13), c.created_at
        assert mkdate(2005, 4, 9, 20, 54, 39), c.closed_at
        assert not c.open
        assert 2 == c.num_changes
        assert 0 == len(c.tags)
        assert -1464925 == c.bounds.top_right.x
        assert 515288620 == c.bounds.top_right.y
        assert -1465242 == c.bounds.bottom_left.x
        assert 515288506 == c.bounds.bottom_left.y
        assert c.type_str() == 'c'
        assert str(c) == 'c34: closed_at=2005-04-09 20:54:39+00:00, bounds=(-0.1465242/51.5288506 -0.1464925/51.5288620), tags={}'
        assert repr(c) == "osmium.osm.Changeset(id=34, uid=1, created_at=datetime.datetime(2005, 4, 9, 19, 54, 13, tzinfo=datetime.timezone.utc), closed_at=datetime.datetime(2005, 4, 9, 20, 54, 39, tzinfo=datetime.timezone.utc), open=False, num_changes=2, bounds=osmium.osm.Box(bottom_left=osmium.osm.Location(x=-1465242, y=515288506), top_right=osmium.osm.Location(x=-1464925, y=515288620)), user='Steve', tags=osmium.osm.TagList({}))"

    assert 1 == area_importer('c34 k2 s2005-04-09T19:54:13Z e2005-04-09T20:54:39Z '
                               'd34 i1 uSteve x-0.1465242 y51.5288506 X-0.1464925 Y51.5288620',
                              changeset=changeset)


def test_entity_operations():
    assert not o.osm.NOTHING
    assert o.osm.NODE

    assert o.osm.AREA | o.osm.NODE | o.osm.WAY | o.osm.RELATION == o.osm.OBJECT
    assert o.osm.ALL & o.osm.RELATION == o.osm.RELATION

    assert ~o.osm.CHANGESET == o.osm.OBJECT



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2022 Sarah Hoffmann.
import pytest
import osmium as o

def test_list_types():
    ml = o.index.map_types()
    assert isinstance(ml, list)
    assert ml


@pytest.fixture
def table():
    return o.index.create_map("flex_mem")

@pytest.mark.parametrize('use_get', [True, False])
@pytest.mark.parametrize('use_set', [True, False])
def test_set_get(table, use_set, use_get):
    if use_set:
        table.set(4, o.osm.Location(3.4, -5.6))
    else:
        table[4] = o.osm.Location(3.4, -5.6)
    if use_get:
        l = table.get(4)
    else:
        l = table[4]
    assert l.lon == pytest.approx(3.4)
    assert l.lat == pytest.approx(-5.6)

def test_get_unset(table):
    with pytest.raises(KeyError):
        table.get(56)

def test_array_get_unset(table):
    with pytest.raises(KeyError):
        table[56]

def test_set_negative(table):
    with pytest.raises(TypeError):
        table.set(-4, o.osm.Location(3.4, -5.6))

def test_array_set_negative(table):
    with pytest.raises(TypeError):
        table[-4] = o.osm.Location(3.4, -5.6)

def test_used_memory(table):
    table.set(4, o.osm.Location(3.4, -5.6))

    assert table.used_memory() > 0

def test_clear(table):
    table.set(593, o.osm.Location(0.35, 45.3))
    table.get(593)
    table.clear()
    with pytest.raises(KeyError):
        table.get(593)



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2022 Sarah Hoffmann.
"""
Tests for all examples.
"""
from pathlib import Path

TEST_DIR = (Path(__file__) / '..').resolve()
TEST_FILE = TEST_DIR / 'example-test.pbf'
TEST_DIFF = TEST_DIR / 'example-test.osc'


def run_example(name, *args):
    filename = (TEST_DIR / '..' / "examples" / (name + ".py")).resolve()
    globvars = dict()

    with filename.open("rb") as f:
        exec(compile(f.read(), str(filename), 'exec'), globvars)

    return globvars['main'](*args)


def test_amenity_list(capsys):
    assert 0 == run_example("amenity_list", TEST_FILE)

    output = capsys.readouterr().out.splitlines()

    assert 835 == len(output)
    assert '1.535245 42.556681 parking' == output[0].strip()
    assert '1.570729 42.529562 parking         Aparcament Comunal' == output[-1].strip()


def test_road_length(capsys):
    assert 0 == run_example("road_length", TEST_FILE)

    output = capsys.readouterr().out.strip()

    assert output == "Total way length: 1590.82 km"


def test_pub_names(capsys):
    assert 0 == run_example("pub_names", TEST_FILE)

    output = capsys.readouterr().out.splitlines()

    assert output == ['Kyu', 'Havana Club', "Mulligan's", 'Bar Broques',
                      'The Camden - English Pub', 'Aspen', 'el Raval']


def test_osm_diff_stats(capsys):
    assert 0 == run_example("osm_diff_stats", TEST_DIFF)

    output = capsys.readouterr().out.splitlines()

    assert output == ['Nodes added: 305',
                     'Nodes modified: 192',
                     'Nodes deleted: 20',
                     'Ways added: 31',
                     'Ways modified: 93',
                     'Ways deleted: 0',
                     'Relations added: 0',
                     'Relations modified: 0',
                     'Relations deleted: 0']


def test_osm_file_stats(capsys):
    assert 0 == run_example("osm_file_stats", TEST_FILE)

    output = capsys.readouterr().out.splitlines()

    assert output == ['Nodes: 211100',
                      'Ways: 10315',
                      'Relations: 244']



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
"""
Version information.
"""

# the major version
pyosmium_major = '4.0'
# current release (Pip version)
pyosmium_release = '4.0.2'

# libosmium version shipped with the Pip release
libosmium_version = '2.20.0'
# protozero version shipped with the Pip release
protozero_version = '1.7.1'
# pybind11 version shipped with the Pip release
pybind11_version = '2.13.6'



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2023 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
""" Helper functions to communicate with replication servers.
"""
from typing import NamedTuple, Optional, Any, Iterator, cast, Dict, Mapping, Tuple
import urllib.request as urlrequest
from urllib.error import URLError
import datetime as dt
from collections import namedtuple
from contextlib import contextmanager
from math import ceil

import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from osmium import MergeInputReader, BaseHandler
from osmium import io as oio
from osmium import version

import logging

LOG = logging.getLogger('pyosmium')
LOG.addHandler(logging.NullHandler())

class OsmosisState(NamedTuple):
    """ Represents a state file of a replication server.
    """
    sequence: int
    "The ID of the replication change on the server."
    timestamp: dt.datetime
    "Date until when changes are contained in the change file."

class DownloadResult(NamedTuple):
    """ Downloaded change.
    """
    id: int
    "The ID of the latest downloaded replication change on the server."
    reader: MergeInputReader
    "[osmium.MergeInputReader][] with all downloaded changes."
    newest: int
    "ID of the newest change available on the server."

class ReplicationServer:
    """ Represents a connection to a  server that publishes replication data.
        Replication change files allow to keep local OSM data up-to-date without
        downloading the full dataset again.

        ReplicationServer may be used as a context manager. In this case, it
        internally keeps a connection to the server making downloads faster.
    """

    def __init__(self, url: str, diff_type: str = 'osc.gz') -> None:
        """ Set up the connection to a replication server.

            `url` contains the base URL of the replication service. This is
            the directory that contains the state file with the current
            state. If the replication service serves something other
            than osc.gz files, set the `diff_type` to the given file suffix.
        """

        self.baseurl = url
        self.diff_type = diff_type
        self.extra_request_params: dict[str, Any] = dict(timeout=60, stream=True)
        self.session: Optional[requests.Session] = None
        self.retry = Retry(total=3, backoff_factor=0.5, allowed_methods={'GET'},
                           status_forcelist=[408, 429, 500, 502, 503, 504])

    def close(self) -> None:
        """ Close any open connection to the replication server.
        """
        if self.session is not None:
            self.session.close()
            self.session = None

    def __enter__(self) -> 'ReplicationServer':
        self.session = requests.Session()
        self.session.mount('http://', HTTPAdapter(max_retries=self.retry))
        self.session.mount('https://', HTTPAdapter(max_retries=self.retry))
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.close()

    def set_request_parameter(self, key: str, value: Any) -> None:
        """ Set a parameter which will be handed to the requests library
            when calling `requests.get()`. This
            may be used to set custom headers, timeouts and similar parameters.
            See the `requests documentation <https://requests.readthedocs.io/en/latest/api/?highlight=get#requests.request>`_
            for possible parameters. Per default, a timeout of 60 sec is set
            and streaming download enabled.
        """
        self.extra_request_params[key] = value

    def make_request(self, url: str) -> urlrequest.Request:
        headers = {"User-Agent" : f"pyosmium/{version.pyosmium_release}"}
        return urlrequest.Request(url, headers=headers)

    def open_url(self, url: urlrequest.Request) -> Any:
        """ Download a resource from the given URL and return a byte sequence
            of the content.
        """
        if 'headers' in self.extra_request_params:
            get_params = self.extra_request_params
        else:
            get_params = dict(self.extra_request_params)
            get_params['headers'] = {k: v for k,v in url.header_items()}

        if self.session is not None:
            return self.session.get(url.get_full_url(), **get_params)

        @contextmanager
        def _get_url_with_session() -> Iterator[requests.Response]:
            with requests.Session() as session:
                session.mount('http://', HTTPAdapter(max_retries=self.retry))
                session.mount('https://', HTTPAdapter(max_retries=self.retry))
                request = session.get(url.get_full_url(), **get_params)
                yield request

        return _get_url_with_session()

    def collect_diffs(self, start_id: int, max_size: int = 1024) -> Optional[DownloadResult]:
        """ Create a MergeInputReader and download diffs starting with sequence
            id `start_id` into it. `max_size`
            restricts the number of diffs that are downloaded. The download
            stops as soon as either a diff cannot be downloaded or the
            unpacked data in memory exceeds `max_size` kB.

            If some data was downloaded, returns a namedtuple with three fields:
            `id` contains the sequence id of the last downloaded diff, `reader`
            contains the MergeInputReader with the data and `newest` is a
            sequence id of the most recent diff available.

            Returns None if there was an error during download or no new
            data was available.
        """
        left_size = max_size * 1024
        current_id = start_id

        # must not read data newer than the published sequence id
        # or we might end up reading partial data
        newest = self.get_state_info()

        if newest is None or current_id > newest.sequence:
            return None

        rd = MergeInputReader()

        while left_size > 0 and current_id <= newest.sequence:
            try:
                diffdata = self.get_diff_block(current_id)
            except:
                LOG.error("Error during diff download. Bailing out.")
                diffdata = ''
            if len(diffdata) == 0:
                if start_id == current_id:
                    return None
                break

            left_size -= rd.add_buffer(diffdata, self.diff_type)
            LOG.debug("Downloaded change %d. (%d kB available in download buffer)",
                      current_id, left_size / 1024)
            current_id += 1

        return DownloadResult(current_id - 1, rd, newest.sequence)

    def apply_diffs(self, handler: BaseHandler, start_id: int,
                    max_size: int = 1024, idx: str = "",
                    simplify: bool = True) -> Optional[int]:
        """ Download diffs starting with sequence id `start_id`, merge them
            together and then apply them to handler `handler`. `max_size`
            restricts the number of diffs that are downloaded. The download
            stops as soon as either a diff cannot be downloaded or the
            unpacked data in memory exceeds `max_size` kB.

            If `idx` is set, a location cache will be created and applied to
            the way nodes. You should be aware that diff files usually do not
            contain the complete set of nodes when a way is modified. That means
            that you cannot just create a new location cache, apply it to a diff
            and expect to get complete way geometries back. Instead you need to
            do an initial data import using a persistent location cache to
            obtain a full set of node locations and then reuse this location
            cache here when applying diffs.

            Diffs may contain multiple versions of the same object when it was
            changed multiple times during the period covered by the diff. If
            `simplify` is set to False then all versions are returned. If it
            is True (the default) then only the most recent version will be
            sent to the handler.

            The function returns the sequence id of the last diff that was
            downloaded or None if the download failed completely.
        """
        diffs = self.collect_diffs(start_id, max_size)

        if diffs is None:
            return None

        diffs.reader.apply(handler, idx=idx, simplify=simplify)

        return diffs.id

    def apply_diffs_to_file(self, infile: str, outfile: str,
                            start_id: int, max_size: int = 1024,
                            set_replication_header: bool = True,
                            extra_headers: Optional[Mapping[str, str]] = None,
                            outformat: Optional[str] = None) -> Optional[Tuple[int, int]]:
        """ Download diffs starting with sequence id `start_id`, merge them
            with the data from the OSM file named `infile` and write the result
            into a file with the name `outfile`. The output file must not yet
            exist.

            `max_size` restricts the number of diffs that are downloaded. The
            download stops as soon as either a diff cannot be downloaded or the
            unpacked data in memory exceeds `max_size` kB.

            If `set_replication_header` is true then the URL of the replication
            server and the sequence id and timestamp of the last diff applied
            will be written into the `writer`. Note that this currently works
            only for the PBF format.

            `extra_headers` is a dict with additional header fields to be set.
            Most notably, the 'generator' can be set this way.

            `outformat` sets the format of the output file. If None, the format
            is determined from the file name.

            The function returns a tuple of last downloaded sequence id and
            newest available sequence id if new data has been written or None
            if no data was available or the download failed completely.
        """
        diffs = self.collect_diffs(start_id, max_size)

        if diffs is None:
            return None

        reader = oio.Reader(infile)
        has_history = reader.header().has_multiple_object_versions

        h = oio.Header()
        h.has_multiple_object_versions = has_history
        if set_replication_header:
            h.set("osmosis_replication_base_url", self.baseurl)
            h.set("osmosis_replication_sequence_number", str(diffs.id))
            info = self.get_state_info(diffs.id)
            if info is not None:
                h.set("osmosis_replication_timestamp", info.timestamp.strftime("%Y-%m-%dT%H:%M:%SZ"))
        if extra_headers is not None:
            for k, v in extra_headers.items():
                h.set(k, v)

        if outformat is None:
            of = oio.File(outfile)
        else:
            of = oio.File(outfile, outformat)

        of.has_multiple_object_versions = has_history
        writer = oio.Writer(of, h)

        LOG.debug("Merging changes into OSM file.")

        diffs.reader.apply_to_reader(reader, writer, has_history)

        reader.close()
        writer.close()

        return (diffs.id, diffs.newest)


    def timestamp_to_sequence(self, timestamp: dt.datetime,
                              balanced_search: bool = False) -> Optional[int]:
        """ Get the sequence number of the replication file that contains the
            given timestamp. The search algorithm is optimised for replication
            servers that publish updates in regular intervals. For servers
            with irregular change file publication dates 'balanced_search`
            should be set to true so that a standard binary search for the
            sequence will be used. The default is good for all known
            OSM replication services.
        """

        # get the current timestamp from the server
        upper = self.get_state_info()

        if upper is None:
            return None
        if timestamp >= upper.timestamp or upper.sequence <= 0:
            return upper.sequence

        # find a state file that is before the required timestamp
        lower = None
        lowerid = 0
        while lower is None:
            LOG.debug("Trying with Id %s", lowerid)
            lower = self.get_state_info(lowerid)

            if lower is not None and lower.timestamp >= timestamp:
                if lower.sequence == 0 or lower.sequence + 1 >= upper.sequence:
                    return lower.sequence
                upper = lower
                lower = None
                lowerid = 0

            if lower is None:
                # no lower yet, so try a higher id (binary search wise)
                newid = int((lowerid + upper.sequence) / 2)
                if newid <= lowerid:
                    # nothing suitable found, so upper is probably the best we can do
                    return upper.sequence
                lowerid = newid

        # Now do a binary search between upper and lower.
        # We could be clever here and compute the most likely state file
        # by interpolating over the timestamps but that creates a whole ton of
        # special cases that need to be handled correctly.
        while True:
            if balanced_search:
                base_splitid = int((lower.sequence + upper.sequence) / 2)
            else:
                ts_int = (upper.timestamp - lower.timestamp).total_seconds()
                seq_int = upper.sequence - lower.sequence
                goal = (timestamp - lower.timestamp).total_seconds()
                base_splitid = lower.sequence + ceil(goal * seq_int / ts_int)
                if base_splitid >= upper.sequence:
                    base_splitid = upper.sequence - 1
            split = self.get_state_info(base_splitid)

            if split is None:
                # file missing, search the next towards lower
                splitid = base_splitid - 1
                while split is None and splitid > lower.sequence:
                    split = self.get_state_info(splitid)
                    splitid -= 1
            if split is None:
                # still nothing? search towards upper
                splitid = base_splitid + 1
                while split is None and splitid < upper.sequence:
                    split = self.get_state_info(splitid)
                    splitid += 1
            if split is None:
                # still nothing? Then lower has to do
                return lower.sequence

            # set new boundary
            if split.timestamp < timestamp:
                lower = split
            else:
                upper = split

            if lower.sequence + 1 >= upper.sequence:
                return lower.sequence


    def get_state_info(self, seq: Optional[int] = None, retries: int = 2) -> Optional[OsmosisState]:
        """ Downloads and returns the state information for the given
            sequence. If the download is successful, a namedtuple with
            `sequence` and `timestamp` is returned, otherwise the function
            returns `None`. `retries` sets the number of times the download
            is retried when pyosmium detects a truncated state file.
        """
        for _ in range(retries + 1):
            ts = None
            next_seq = None

            try:
                with self.open_url(self.make_request(self.get_state_url(seq))) as response:
                    if hasattr(response, 'iter_lines'):
                        # generated by requests
                        response.raise_for_status()
                        lines = response.iter_lines()
                    else:
                        lines = response

                    for line in lines:
                        line = line.decode('utf-8').strip()
                        if '#' in line:
                            line = line[0:line.index('#')]
                        else:
                            line = line.strip()
                        if line:
                            kv = line.split('=', 2)
                            if len(kv) != 2:
                                return None
                            if kv[0] == 'sequenceNumber':
                                next_seq = int(kv[1])
                            elif kv[0] == 'timestamp':
                                try:
                                    ts = dt.datetime.strptime(kv[1], "%Y-%m-%dT%H\\:%M\\:%SZ")
                                except ValueError:
                                    break
                                ts = ts.replace(tzinfo=dt.timezone.utc)

            except (URLError, IOError) as err:
                LOG.debug("Loading state info failed with: %s", str(err))
                return None

            if ts is not None and next_seq is not None:
                return OsmosisState(sequence=next_seq, timestamp=ts)

        return None

    def get_diff_block(self, seq: int) -> str:
        """ Downloads the diff with the given sequence number and returns
            it as a byte sequence. Throws an :code:`requests.HTTPError`
            if the file cannot be downloaded.
        """
        with self.open_url(self.make_request(self.get_diff_url(seq))) as resp:
            if hasattr(resp, 'content'):
                # generated by requests
                resp.raise_for_status()
                return cast(str, resp.content)

            # generated by urllib.request
            return cast(str, resp.read())


    def get_state_url(self, seq: Optional[int]) -> str:
        """ Returns the URL of the state.txt files for a given sequence id.

            If seq is `None` the URL for the latest state info is returned,
            i.e. the state file in the root directory of the replication
            service.
        """
        if seq is None:
            return self.baseurl + '/state.txt'

        return '%s/%03i/%03i/%03i.state.txt' % \
               (self.baseurl, seq / 1000000, (seq % 1000000) / 1000, seq % 1000)


    def get_diff_url(self, seq: int) -> str:
        """ Returns the URL to the diff file for the given sequence id.
        """
        return '%s/%03i/%03i/%03i.%s' % \
               (self.baseurl,
                seq / 1000000, (seq % 1000000) / 1000, seq % 1000,
                self.diff_type)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2023 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import Optional, Union, Any, Mapping, Sequence, Tuple, TYPE_CHECKING
from datetime import datetime

if TYPE_CHECKING:
    import osmium.osm

    OSMObjectLike = Union['OSMObject', osmium.osm.OSMObject[Any]]

    TagSequence = Union[osmium.osm.TagList, Mapping[str, str], Sequence[Tuple[str, str]]]
    LocationLike = Union[osmium.osm.Location, Tuple[float, float]]
    NodeSequence = Union[osmium.osm.NodeRefList, Sequence[Union[osmium.osm.NodeRef, int]]]
    MemberSequence = Union[osmium.osm.RelationMemberList,
                       Sequence[Union[osmium.osm.RelationMember, Tuple[str, int, str]]]]

class OSMObject:
    """ Mutable version of [osmium.osm.OSMObject][].
    """

    def __init__(self, base: Optional['OSMObjectLike'] = None,
                 id: Optional[int] = None, version: Optional[int] = None,
                 visible: Optional[bool] = None, changeset: Optional[int] = None,
                 timestamp: Optional[datetime] = None, uid: Optional[int] = None,
                 tags: Optional['TagSequence'] = None, user: Optional[str] = None) -> None:
        """ Initialise an object with the following optional
            attributes: `id`, `version`, `visible`, `changeset`, `timestamp`,
            `uid` and `tags`. Timestamps may be strings or datetime objects.
            Tags can be an osmium.osm.TagList, a dict-like object
            or a list of tuples, where each tuple contains a (key value) string pair.

            If the `base` parameter is given in the constructor, then the object
            will be initialised first from the attributes of this base object.
        """
        if base is None:
            self.id = id
            self.version = version
            self.visible = visible
            self.changeset = changeset
            self.timestamp = timestamp
            self.uid = uid
            self.tags = tags
            self.user = user
        else:
            self.id = base.id if id is None else id
            self.version = base.version if version is None else version
            self.visible = base.visible if visible is None else visible
            self.changeset = base.changeset if changeset is None else changeset
            self.timestamp = base.timestamp if timestamp is None else timestamp
            self.uid = base.uid if uid is None else uid
            self.tags = base.tags if tags is None else tags
            self.user = base.user if user is None else user


class Node(OSMObject):
    """ The mutable version of [osmium.osm.Node][].
    """

    def __init__(self, base: Optional[Union['Node', 'osmium.osm.Node']] = None,
                 location: Optional['LocationLike'] = None,
                 **attrs: Any) -> None:
        """ Initialise a node with all optional attributes
            from osmium.osm.mutable.OSMObject as well as a `location` attribute.
            This may either be an [osmium.osm.Location][] or a tuple of
            lon/lat coordinates.

        """
        OSMObject.__init__(self, base=base, **attrs)
        if base is None:
            self.location = location
        else:
            self.location = location if location is not None else base.location


class Way(OSMObject):
    """ The mutable version of [osmium.osm.Way][].
    """

    def __init__(self, base: Optional[Union['Way', 'osmium.osm.Way']] = None,
                 nodes: Optional['NodeSequence'] = None, **attrs: Any) -> None:
        """ Initialise a way with all optional attributes
            from osmium.osm.mutable.OSMObject as well as a `nodes` attribute.
            This may either be an [osmium.osm.NodeRefList][] or a list
            consisting of [osmium.osm.NodeRef][] or simple node ids.
        """
        OSMObject.__init__(self, base=base, **attrs)
        if base is None:
            self.nodes = nodes
        else:
            self.nodes = nodes if nodes is not None else base.nodes

class Relation(OSMObject):
    """ The mutable version of [osmium.osm.Relation][].
    """

    def __init__(self, base: Optional[Union['Relation', 'osmium.osm.Relation']] = None,
                 members: Optional['MemberSequence'] = None, **attrs: Any) -> None:
        """ Initialise a relation with all optional attributes
            from osmium.osm.mutable.OSMObject as well as a `members` attribute.
            This may either be an [osmium.osm.RelationMemberList][] or
            a list consisting of [osmium.osm.RelationMember][] or
            tuples of (type, id, role). The
            member type must be a single character 'n', 'w' or 'r'.
        """
        OSMObject.__init__(self, base=base, **attrs)
        if base is None:
            self.members = members
        else:
            self.members = members if members is not None else base.members


def create_mutable_node(node: Union[Node, 'osmium.osm.Node'], **args: Any) -> Node:
    """ Create a mutable node replacing the properties given in the
        named parameters. Note that this function only creates a shallow
        copy which is still bound to the scope of the original object.
    """
    return Node(base=node, **args)

def create_mutable_way(way: Union[Way, 'osmium.osm.Way'], **args: Any) -> Way:
    """ Create a mutable way replacing the properties given in the
        named parameters. Note that this function only creates a shallow
        copy which is still bound to the scope of the original object.
    """
    return Way(base=way, **args)

def create_mutable_relation(rel: Union[Relation, 'osmium.osm.Relation'], **args: Any) -> Relation:
    """ Create a mutable relation replacing the properties given in the
        named parameters. Note that this function only creates a shallow
        copy which is still bound to the scope of the original object.
    """
    return Relation(base=rel, **args)




# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2024 Sarah Hoffmann.
import pytest
import osmium as o


def test_read_node_location_with_handler(opl_reader):
    idx = o.index.create_map("flex_mem")
    hdlr = o.NodeLocationsForWays(idx)

    data = """\
           n1 x6 y7
           n45 x-3 y0
           """

    o.apply(opl_reader(data), hdlr)

    assert idx.get(1).lon == pytest.approx(6)
    assert idx.get(1).lat == pytest.approx(7)
    assert idx.get(45).lon == pytest.approx(-3)
    assert idx.get(45).lat == 0.0

    with pytest.raises(KeyError):
        idx.get(2)


@pytest.mark.parametrize('ignore_error', [(True, False)])
def test_apply_node_location_handler(opl_reader, ignore_error):

    hdlr = o.NodeLocationsForWays(o.index.create_map("flex_mem"))
    if ignore_error:
        hdlr.ignore_errors()

    class WayNodeHandler:
        def __init__(self):
            self.collect = []
            self.with_error = []

        def way(self, w):
            try:
                self.collect.append((w.id, [(n.lon, n.lat) for n in w.nodes]))
            except o.InvalidLocationError:
                self.with_error.append(w.id)


    data = """\
           n1 x6 y7
           n2 x6 y7.1
           n45 x-3 y0
           n55 x-2.9 y0
           w3 Nn1,n2
           w4 Nn45,n55,n56
           """

    tester = WayNodeHandler()

    if ignore_error:
        o.apply(opl_reader(data), hdlr, tester)

        assert tester.collect == [(3, [(pytest.approx(6), pytest.approx(7)),
                                      (pytest.approx(6), pytest.approx(7.1))])]
        assert tester.with_error == [4]
    else:
        with pytest.raises(osmium.InvalidLocationError):
            o.apply(opl.reader(data), hdlr, tester)


def test_apply_invalid_handler_object(opl_reader):
    class DummyHandler:
        def some_func():
            print('A')

    with pytest.raises(TypeError):
        o.apply(opl_reader("n1 x2 z4"), DummyHandler())


def test_mixed_handlers(opl_reader):
    logged = []

    class OldStyle(o.SimpleHandler):
        def node(self, n):
            logged.append('old')

    class NewStyle:
        def node(self, n):
            logged.append('new')

    o.apply(opl_reader("n1 x0 y0"), NewStyle(), OldStyle(), NewStyle(), OldStyle())

    assert logged == ['new', 'old', 'new', 'old']


def test_value_propagation(opl_reader):
    logged = []

    class FirstHandler:
        def node(self, n):
            n.saved = 45674

    class SecondHandler:
        def node(self, n):
            logged.append(n.saved)

    o.apply(opl_reader("n1 x0 y0"), FirstHandler(), SecondHandler())

    assert logged == [45674]



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2022 Sarah Hoffmann.
import pytest

import osmium as o

from helpers import CountingHandler

class NullHandler:

    def node(self, n):
        pass

def _run_file(fn):
    with o.io.Reader(fn) as rd:
        o.apply(rd, NullHandler())


@pytest.mark.parametrize('as_string', [True, False])
def test_file_simple(tmp_path, as_string):
    fn = tmp_path / 'text.opl'
    fn.write_text('n1')

    if as_string:
        fn = str(fn)

    for n in o.FileProcessor(o.io.File(fn)):
        assert n.is_node()
        assert n.id == 1


@pytest.mark.parametrize('as_string', [True, False])
def test_file_with_format(tmp_path, as_string):
    fn = tmp_path / 'text.txt'
    fn.write_text('n1')

    if as_string:
        fn = str(fn)

    for n in o.FileProcessor(o.io.File(fn, 'opl')):
        assert n.is_node()
        assert n.id == 1


def test_node_only(test_data):
    _run_file(test_data('n1'))


def test_way_only(test_data):
    _run_file(test_data('w1 Nn1,n2,n3'))


def test_relation_only(test_data):
    _run_file(test_data('r573 Mw1@'))


def test_node_with_tags(test_data):
    _run_file(test_data('n32 Tbar=xx'))


def test_way_with_tags(test_data):
    _run_file(test_data('w5666 Nn1,n2,n3 Tbar=xx'))


def test_relation_with_tags(test_data):
    _run_file(test_data('r573 Mw1@ Tbar=xx'))


def test_broken_timestamp(test_data):
    fn = test_data('n1 tx')

    with o.io.Reader(fn) as rd:
        with pytest.raises(RuntimeError):
            o.apply(rd, NullHandler())


@pytest.mark.parametrize('as_string', [True, False])
def test_file_header(tmp_path, as_string):
    fn = tmp_path / 'empty.xml'
    fn.write_text("""<?xml version='1.0' encoding='UTF-8'?>
    <osm version="0.6" generator="test-pyosmium" timestamp="2014-08-26T20:22:02Z">
         <bounds minlat="-90" minlon="-180" maxlat="90" maxlon="180"/>
    </osm>
    """)

    if as_string:
        fn = str(fn)

    with o.io.Reader(fn) as rd:
        h = rd.header()
        assert not h.has_multiple_object_versions
        assert h.box().valid()
        assert h.box().size() == 64800.0


def test_reader_with_filebuffer():
    rd = o.io.Reader(o.io.FileBuffer('n1 x4 y1'.encode('utf-8'), 'opl'))
    handler = CountingHandler()

    o.apply(rd, handler)

    assert handler.counts == [1, 0, 0, 0]



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2023 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.

from ._replication import (newest_change_from_file as newest_change_from_file)

from .server import (ReplicationServer as ReplicationServer,
                     OsmosisState as OsmosisState,
                     DownloadResult as DownloadResult)
from .utils import (ReplicationHeader as ReplicationHeader,
                    get_replication_header as get_replication_header)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import Any, Optional, Union
from pathlib import Path
from tempfile import TemporaryDirectory
import os

from osmium._osmium import SimpleWriter
from osmium import IdTracker
from osmium.io import File, FileBuffer
from osmium.file_processor import FileProcessor, zip_processors

class ForwardReferenceWriter:
    """ Writer that adds forward-referenced objects optionally also making
        the final file reference complete. An object is a forward reference
        when it directly or indirectly needs one of the objects originally
        written out.

        The collected data is first written into a temporary file, When the
        writer is closed, the references are collected from the reference file
        and written out together with the collected data into the final file.

        The writer should usually be used as a context manager.
    """

    def __init__(self, outfile: Union[str, 'os.PathLike[str]', File],
                 ref_src: Union[str, 'os.PathLike[str]', File, FileBuffer],
                 overwrite: bool=False, back_references: bool=True,
                 remove_tags: bool=True, forward_relation_depth: int=0,
                 backward_relation_depth: int=1) -> None:
        """ Create a new writer.

            `outfile` is the name of the output file to write. The file must
            not yet exist unless `overwrite` is set to True.

            `ref_src` is the OSM input file, where to take the reference objects
            from. This is usually the same file the data to be written is taken
            from.

            The writer will collect back-references by default to make the
            file reference-complete. Set `back_references=False` to disable
            this behaviour.

            The writer will not complete nested relations by default. If you
            need nested relations, set `relation_depth` to the minimum depth
            to which relations shall be completed.
        """
        self.outfile = outfile
        self.tmpdir: Optional['TemporaryDirectory[Any]'] = TemporaryDirectory()
        self.writer = SimpleWriter(str(Path(self.tmpdir.name, 'forward_writer.osm.pbf')))
        self.overwrite = overwrite
        self.back_references = back_references
        self.id_tracker = IdTracker()
        self.ref_src = ref_src
        self.forward_relation_depth = forward_relation_depth
        self.backward_relation_depth = backward_relation_depth


    def __enter__(self) -> 'ForwardReferenceWriter':
        return self


    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        if exc_type is None:
            self.close()
        else:
            # Exception occured. Do not write out the final file.
            self.writer.close()


    def add(self, obj: Any) -> None:
        """ Write an arbitrary OSM object. This can be either an
            osmium object or a Python object that has the appropriate
            attributes.
        """
        if hasattr(obj, 'location'):
            self.id_tracker.add_node(obj.id)
        elif hasattr(obj, 'nodes'):
            self.id_tracker.add_way(obj.id)
        elif hasattr(obj, 'members'):
            self.id_tracker.add_relation(obj.id)
        self.writer.add(obj)


    def add_node(self, n: Any) -> None:
        """ Write out an OSM node.
        """
        self.id_tracker.add_node(n.id)
        self.writer.add_node(n)


    def add_way(self, w: Any) -> None:
        """ Write out an OSM way.
        """
        self.id_tracker.add_way(w.id)
        self.writer.add_way(w)


    def add_relation(self, r: Any) -> None:
        """ Write out an OSM relation.
        """
        self.id_tracker.add_relation(r.id)
        self.writer.add_relation(r)


    def close(self) -> None:
        """ Close the writer and write out the final file.

            The function will be automatically called when the writer
            is used as a context manager.
        """
        if self.tmpdir is not None:
            self.writer.close()

            self.id_tracker.complete_forward_references(self.ref_src,
                                                        relation_depth=self.forward_relation_depth)
            if self.back_references:
                self.id_tracker.complete_backward_references(self.ref_src,
                                                             relation_depth=self.backward_relation_depth)

            fp1 = FileProcessor(Path(self.tmpdir.name, 'forward_writer.osm.pbf'))
            fp2 = FileProcessor(self.ref_src).with_filter(self.id_tracker.id_filter())


            with SimpleWriter(self.outfile, overwrite=self.overwrite) as writer:
                for o1, o2 in zip_processors(fp1, fp2):
                    if o1:
                        writer.add(o1)
                    else:
                        writer.add(o2)

            self.tmpdir.cleanup()
            self.tmpdir = None





# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.

from . import BaseHandler, BufferIterator
from ._osmium import HandlerLike

class AreaManagerSecondPassHandler(BaseHandler): ...

class AreaManagerBufferHandler(BaseHandler):...


class AreaManager(BaseHandler):
    """ Handler class that manages building area objects from ways
        and relations.

        Building area objects always requires two passes through the file:
        in the first pass, the area manager collects the relation candidates
        for areas and IDs of all ways that are needed to build their areas.
        During the second pass of the file the areas are assembled: areas
        from ways are created immediately when the handler encounters a
        closed way. Areas for relations are built as soon as all the ways
        that the relation needs are available.

        You usually should not be using the AreaManager direcly. The
        interface of the handler is considered an internal implementation
        detail and may change in future versions of pyosmium. Area assembly
        can be enabled through the [SimpleHandler][osmium.SimpleHandler]
        and the [FileProcessor][osmium.FileProcessor].
    """
    def __init__(self) -> None:
        """ Set up a new area manager.
        """
    def first_pass_handler(self) -> 'AreaManager':
        """ Return a handler object to be used for the first pass through
            a file. It collects information about area relations and their
            way members.
        """
    def second_pass_handler(self, *handlers: HandlerLike) -> AreaManagerSecondPassHandler:
        """ Return a handler used for the second pass of the
            file, where areas are assembled. Pass the chain of filters and
            handlers that should be applied the areas.
        """
    def second_pass_to_buffer(self, callback: BufferIterator) -> AreaManagerBufferHandler:
        """ Return a handler for the second pass of the file, which stores
            assembled areas in the given buffer.
        """



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2022 Sarah Hoffmann.
import pytest

import osmium as o

def test_waynode_length(simple_handler):
    data = """\
           w593
           w4 Nn1,n2,n-34
           w8 Nn12,n12,n12,n0
           """

    lens = {}
    def way(w):
        lens[w.id] = len(w.nodes)

    simple_handler(data, way=way)

    assert lens == { 593 : 0, 4 : 3, 8 : 4 }


def test_node_ids(simple_handler):
    refs = []
    def way(w):
        refs.extend(n.ref for n in w.nodes)
        assert w.nodes[-2].ref == -34
        with pytest.raises(IndexError):
            w.nodes[5]
        with pytest.raises(IndexError):
            w.nodes[-6]

    simple_handler("w4 Nn1,n1,n34359737784,n-34,n0", way=way)

    assert refs == [1, 1, 34359737784, -34, 0]


def test_missing_location_without_location_handler(simple_handler):
    data = """\
           n1 x0.5 y10.0
           w4 Nn1
           """

    refs = []
    def way(w):
        refs.extend(n.ref for n in w.nodes)
        assert not w.nodes[0].location.valid()
        with pytest.raises(o.InvalidLocationError):
            w.nodes[0].location.lat
        with pytest.raises(o.InvalidLocationError):
            w.nodes[0].location.lon

    simple_handler(data, way=way)

    assert refs == [1]


def test_valid_locations(simple_handler):
    data = """\
           n1 x0.5 y10.0
           w4 Nn1
           """

    locations = []
    def way(w):
        assert all(n.location.valid() for n in w.nodes)
        locations.extend((int(10 * n.location.lon), int(10 * n.location.lat))
                         for n in w.nodes)

    simple_handler(data, way=way, locations=True)

    assert locations == [(5, 100)]



"""
Search for pubs in an osm file and list their names.
"""
import osmium
import sys

def main(osmfile):
    for obj in osmium.FileProcessor(osmfile)\
                     .with_filter(osmium.filter.KeyFilter('amenity'))\
                     .with_filter(osmium.filter.KeyFilter('name')):
        if obj.tags['amenity'] == 'pub':
            print(obj.tags['name'])

    return 0

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: python %s <osmfile>" % sys.argv[0])
        sys.exit(-1)

    exit(main(sys.argv[1]))



"""
Compute the total length of highways in an osm file.

Shows how to extract the geometry of a way.
"""
import osmium as o
import sys

def main(osmfile):
    total = 0.0
    # As we need the way geometry, the node locations need to be cached.
    # This is enabled with the with_locations() function.
    for obj in o.FileProcessor(osmfile, o.osm.NODE | o.osm.WAY)\
                .with_locations()\
                .with_filter(o.filter.KeyFilter('highway')):
        if obj.is_way():
            try:
                total += o.geom.haversine_distance(obj.nodes)
            except o.InvalidLocationError:
                # A location error might occur if the osm file is an extract
                # where nodes of ways near the boundary are missing.
                print("WARNING: way %d incomplete. Ignoring." % obj.id)

    print('Total way length: %.2f km' % (total/1000))

    return 0

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: python %s <osmfile>" % sys.argv[0])
        sys.exit(-1)

    exit(main(sys.argv[1]))



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import os
import re
import sys
import platform
import subprocess
import urllib.request
import tarfile
from pathlib import Path

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist as orig_sdist

BASEDIR = os.path.split(os.path.abspath(__file__))[0]

class PyosmiumSDist(orig_sdist):

    contrib = (
        ('libosmium', 'https://github.com/osmcode/libosmium/archive/v{}.tar.gz'),
        ('protozero', 'https://github.com/mapbox/protozero/archive/v{}.tar.gz'),
        ('pybind11', 'https://github.com/pybind/pybind11/archive/v{}.tar.gz'),
    )

    def make_release_tree(self, base_dir, files):
        orig_sdist.make_release_tree(self, base_dir, files)

        # add additional dependecies in the required version
        for name, tar_src in self.contrib:
            tarball = tar_src.format(versions[name + '_version'])
            print("Downloading and adding {} sources from {}".format(name, tarball))
            base = Path("-".join((name, versions[name + '_version'])))
            dest = Path(base_dir) / "contrib" / name
            with urllib.request.urlopen(tarball) as reader:
                with tarfile.open(fileobj=reader, mode='r|gz') as tf:
                    for member in tf:
                        fname = Path(member.name)
                        if not fname.is_absolute():
                            fname = fname.relative_to(base)
                            if member.isdir():
                                (dest / fname).mkdir(parents=True, exist_ok=True)
                            elif member.isfile():
                                with tf.extractfile(member) as memberfile:
                                    with (dest / fname).open('wb') as of:
                                        of.write(memberfile.read())



def get_versions():
    """ Read the version file.

        The file cannot be directly imported because it is not installed
        yet.
    """
    version_py = os.path.join(BASEDIR, "src/osmium/version.py")
    v = {}
    with open(version_py) as version_file:
        # Execute the code in version.py.
        exec(compile(version_file.read(), version_py, 'exec'), v)

    return v

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            nbr_cpus = os.cpu_count() or 2 # fallback if None is returned
            build_args += ['--', f'-j{nbr_cpus}']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())

        if 'LIBOSMIUM_PREFIX' in env:
            cmake_args += ['-DOSMIUM_INCLUDE_DIR={}/include'.format(env['LIBOSMIUM_PREFIX'])]
        elif os.path.exists(os.path.join(BASEDIR, 'contrib', 'libosmium', 'include', 'osmium', 'version.hpp')):
            cmake_args += ['-DOSMIUM_INCLUDE_DIR={}/contrib/libosmium/include'.format(BASEDIR)]

        if 'PROTOZERO_PREFIX' in env:
            cmake_args += ['-DPROTOZERO_INCLUDE_DIR={}/include'.format(env['PROTOZERO_PREFIX'])]
        elif os.path.exists(os.path.join(BASEDIR, 'contrib', 'protozero', 'include', 'protozero', 'version.hpp')):
            cmake_args += ['-DPROTOZERO_INCLUDE_DIR={}/contrib/protozero/include'.format(BASEDIR)]

        if 'PYBIND11_PREFIX' in env:
            cmake_args += ['-DPYBIND11_PREFIX={}'.format(env['PYBIND11_PREFIX'])]
        elif os.path.exists(os.path.join(BASEDIR, 'contrib', 'pybind11')):
            cmake_args += ['-DPYBIND11_PREFIX={}/contrib/pybind11'.format(BASEDIR)]

        if 'BOOST_PREFIX' in env:
            cmake_args += ['-DBOOST_ROOT={}'.format(env['BOOST_PREFIX'])]

        if 'CMAKE_CXX_STANDARD' in env:
            cmake_args += ['-DCMAKE_CXX_STANDARD={}'.format(env['CMAKE_CXX_STANDARD'])]

        cmake_args += [f"-DWITH_LZ4={env.get('CMAKE_WITH_LZ4', 'ON')}"]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)

versions = get_versions()

setup(
    scripts=['tools/pyosmium-get-changes', 'tools/pyosmium-up-to-date'],
    ext_modules=[CMakeExtension('cmake_example')],
    cmdclass=dict(build_ext=CMakeBuild, sdist=PyosmiumSDist),
    zip_safe=False,
)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest

import osmium as o

from helpers import IDCollector

def test_id_filter_bad_argument():
    with pytest.raises(TypeError):
        o.filter.IdFilter(None)


def test_id_filter_simple(opl_reader):
    data = """\
           n1
           n2
           n4
           n200
           """

    ids = IDCollector()

    o.apply(opl_reader(data), o.filter.IdFilter([2, 5, 200, 201]), ids)

    assert ids.nodes == [2, 200]



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2024 Sarah Hoffmann.
from pathlib import Path

import osmium as o

from helpers import CountingHandler

TEST_FILE = str((Path(__file__) / '..' / 'example-test.pbf').resolve())

def test_area_handler():
    area = o.area.AreaManager()

    o.apply(o.io.Reader(TEST_FILE), area.first_pass_handler())

    ch_area = CountingHandler()
    ch_others = CountingHandler()

    lh = o.NodeLocationsForWays(o.index.create_map("flex_mem"))
    lh.ignore_errors()

    o.apply(o.io.Reader(TEST_FILE), lh,
            ch_others, area.second_pass_handler(ch_area))

    assert ch_area.counts == [0, 0, 0, 5239]
    assert ch_others.counts == [211100, 10315, 244, 0]


def test_area_buffer_handler():
    area = o.area.AreaManager()

    o.apply(o.io.Reader(TEST_FILE), area.first_pass_handler())

    lh = o.NodeLocationsForWays(o.index.create_map("flex_mem"))
    lh.ignore_errors()

    buf = o.BufferIterator()

    o.apply(o.io.Reader(TEST_FILE), lh, area.second_pass_to_buffer(buf))

    counts = 0
    for obj in buf:
        assert obj.is_area()
        counts += 1

    assert counts == 5239



# SPDX-License-Identifier: BSD
#
# This file is part of Pyosmium.
#
# Copyright (C) 2022 Sarah Hoffmann.

import osmium.replication.utils as rutil

from helpers import mkdate

def test_get_replication_header_empty(tmp_path):
    fn = tmp_path / 'test.opl'
    fn.write_text('n6365 v1 c63965061 t2018-10-29T03:56:07Z i8369524 ux x1 y7')

    val = rutil.get_replication_header(str(fn))
    assert val.url is None
    assert val.sequence is None
    assert val.timestamp is None


def test_get_replication_header_full(test_data_dir):
    val = rutil.get_replication_header(str(test_data_dir / 'example-test.pbf'))

    assert val.url == 'http://download.geofabrik.de/europe/andorra-updates'
    assert val.sequence == 2167
    assert val.timestamp == mkdate(2019, 2, 23, 21, 15, 2)
